# -*- coding: utf-8 -*-

# Copyright © 2020-2021, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from ..attribute import Initializer
from ..artefact import Analysis
from .abc import *
from .plugin import *
from . import models

class MapperInitializer(Initializer, MapperAttribute):
    __slots__ = ()
    def from_plugin(self, plugin, **kwargs):
        self.specialize( MapperPlugin, plugin, **kwargs )
    def from_callable(self, cls, **kwargs):
        self.from_plugin(cls, **kwargs)
    def from_maps(self, label_or_maps=None, exclude_attrs=(), verbose=False):
        """
        Sets the mapper up on basis of parameters in a Maps object.

        If a label is passed instead of a Maps object (or nothing),
        the corresponding artefact is sought for in the analysis tree,
        starting from the first branch (first sampling) in the first
        tree.

        This method has been designed for the *stochastic.dv* plugin.
        Application to other plugins has not been tested.

        Arguments:

            label_or_maps (*Maps* or *Analysis* or *int* or *str* or *tuple* or *list*):
                maps object or label(s) of the maps artefact to be found in any
                registered analysis tree

            exclude_attrs (iterable):
                attributes of the maps object to be ignored

            verbose (bool):
                prints the plugin name and loaded attributes

        """
        if label_or_maps is None or isinstance(label_or_maps, (int, str)):
            label = label_or_maps
            sampling = first(self._eldest_parent.roi.as_support_regions()).get_sampling()
            maps = sampling.get_child(label)
        elif isinstance(label_or_maps, (tuple, list)):
            sampling_label, labels = label_or_maps[0], label_or_maps[1:]
            sampling = first(self._eldest_parent.roi.as_support_regions()).get_sampling(sampling_label)
            #
            node = sampling
            for label in labels:
                node = node.get_child(label)
            maps = node
        else:
            maps = label_or_maps
        if isinstance(maps, Analysis):
            maps = maps.data
        plugin = maps.mode
        if plugin is None:
            raise ValueError('undefined plugin name')
        self.from_plugin(plugin)
        self = self._eldest_parent.mapper
        exclude_attrs = list(exclude_attrs) # copy
        exclude_attrs += ['mode', 'runtime', 'posteriors', 'niter', 'resolution', 'sigma']
        attrs = []
        for attr in maps.__dict__:
            if not (attr[0] == '_' or attr in exclude_attrs):
                val = getattr(maps, attr)
                if val is not None:
                    setattr(self, attr, val)
                    attrs.append((attr, val))
        if verbose:
            logger = self.logger
            if attrs:
                logger.info("loading plugin: '{}'\nwith parameters:\n - {}".format(
                    plugin,
                    "\n - ".join([
                        ("{}: '{}'" if isinstance(val, str) else "{}: {}").format(key, val) \
                                for key, val in attrs
                    ])))
            else:
                logger.info("plugin:\t'{}'\nwith no parameters".format(plugin))

    @property
    def _mpl_impl(self):
        from .mpl import Mpl
        return Mpl
    @property
    def mpl(self):
        """ tramway.analyzer.mapper.mpl.Mpl: Matplotlib utilities """
        return self._mpl_impl(self)

    @property
    def _plotly_impl(self):
        from .plotly import Plotly
        return Plotly
    @property
    def plotly(self):
        """ tramway.analyzer.mapper.plotly.Plotly: Plotly utilities """
        return self._plotly_impl(self)

__all__ = [ 'Mapper', 'MapperInitializer', 'MapperPlugin', 'models' ]




# -*- coding: utf-8 -*-

# Copyright © 2020-2021, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from ..attribute import *
from ..artefact import analysis
from .. import attribute
from .abc import *
from tramway.inference import plugins
from tramway.helper.inference import Infer

class MapperAttribute(object):
    """ Mixin class for accessing the `localization_precision`,
    `localization_error` and `temperature` attributes.
    """
    __slots__ = ()

    @property
    def localization_precision(self):
        return self._parent.localization_precision
    @localization_precision.setter
    def localization_precision(self, sigma):
        self._parent.localization_precision = sigma

    @property
    def localization_error(self):
        return self._parent.localization_error
    @localization_error.setter
    def localization_error(self, sigma2):
        self._parent.localization_error = sigma2

    @property
    def temperature(self):
        return self._parent.temperature
    @temperature.setter
    def temperature(self, T):
        self._parent.temperature = T


class MapperPlugin(AnalyzerNode, MapperAttribute):
    __slots__ = ('_name','_module','_setup','_mapper','_kwargs')
    def __init__(self, plugin, **kwargs):
        init_kwargs = {}
        for k in attribute.__analyzer_node_init_args__:
            try:
                arg = kwargs.pop(k)
            except KeyError:
                pass
            else:
                init_kwargs[k] = arg
        AnalyzerNode.__init__(self, **init_kwargs)
        self._name = self._module = self._mapper = None
        self._setup, self._kwargs = {}, {}
        if isinstance(plugin, str):
            self._name = plugin
            try:
                plugin = plugins[plugin]
            except KeyError:
                raise KeyError('no such plugin: {}'.format(self._name))
        if callable(plugin):
            self._setup = kwargs
            self._mapper = plugin
        elif isinstance(plugin, tuple):
            setup, self._module = plugin
            self._setup = dict(setup) # copy
            self._mapper = getattr(self._module, self._setup.pop('infer'))
    @property
    def name(self):
        return self._name
    @property
    def setup(self):
        return self._setup
    def get_plugin_arg(self, attrname):
        try:
            return self._setup[attrname]
        except KeyError:
            return self._kwargs.get(attrname, None)
    def __getattr__(self, attrname):
        return self.get_plugin_arg(attrname)
    def set_plugin_arg(self, attrname, val):
        if attrname in self._setup:
            self._setup[attrname] = val
        else:
            self._kwargs[attrname] = val
            if attrname.endswith('time_prior'):
                self.time.enable_regularization()
    def __setattr__(self, attrname, val):
        try:
            AnalyzerNode.__setattr__(self, attrname, val)
        except AttributeError:
            self.set_plugin_arg(attrname, val)
    @analysis
    def infer(self, sampling):
        helper = Infer()
        helper.prepare_data(sampling)
        distr_kwargs, infer_kwargs = {}, {}
        for k in self._kwargs:
            if k in ('new_cell','new_group','include_empty_cells','grad','rgrad','cell_sampling'):
                distr_kwargs[k] = self._kwargs[k]
            if k not in ('new_cell','new_group','include_empty_cells','grad','cell_sampling'):
                infer_kwargs[k] = self._kwargs[k]
        infer_kwargs['sigma'] = self._parent.spt_data.localization_precision
        if 'cell_sampling' not in distr_kwargs and \
                self.time.initialized and not self.time.regularize_in_time:
            distr_kwargs['cell_sampling'] = 'connected'
        cells = helper.distribute(**distr_kwargs)
        helper.name, helper.setup, helper._infer = self.name, self.setup, self._mapper
        cells = helper.overload_cells(cells)
        maps = helper.infer(cells, **infer_kwargs)
        return maps
    @property
    def time(self):
        return self._parent.time

    @property
    def _mpl_impl(self):
        from .mpl import Mpl
        return Mpl
    @property
    def mpl(self):
        """ tramway.analyzer.mapper.mpl.Mpl: Matplotlib utilities """
        return self._mpl_impl(self)

    @property
    def _plotly_impl(self):
        from .plotly import Plotly
        return Plotly
    @property
    def plotly(self):
        """ tramway.analyzer.mapper.plotly.Plotly: Plotly utilities """
        return self._plotly_impl(self)

Mapper.register(MapperPlugin)




# -*- coding: utf-8 -*-

# Copyright © 2018, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from math import *
from .base import Voronoi
import numpy as np
import pandas as pd
import scipy.sparse as sparse
from scipy.optimize import minimize_scalar
from collections import OrderedDict
from warnings import warn


class HexagonalMesh(Voronoi):
    """2D hexagonal mesh.

    Attributes:
        tilt (float):
            angle of the "main" axis, in pi/6 radians;
            0= the main axis is the x-axis; 1= the main axis is the y-axis.
        hexagon_radius (float): hexagon radius.
        hexagon_count (pair of ints):
            number of hexagons along the "main" axis,
            number of hexagons along the perpendical axis.
        min_probability (float): (not used)
        avg_probability (float):
            inverse of the average number of points per cell.
        max_probability (float): (not used)
        min_distance (float):
            minimum distance between adjacent cell centers;
            ignored if `avg_distance` is defined.
        avg_distance (float):
            average distance between adjacent cell centers;
            ignored if `avg_probability` is defined.

    """

    __slots__ = ('tilt', 'hexagon_radius', 'hexagon_count',
        'min_probability', 'avg_probability', 'max_probability',
        'min_distance', 'avg_distance')

    def __init__(self, scaler=None, tilt=0., \
        min_probability=None, max_probability=None, avg_probability=None, \
        min_distance=None, avg_distance=None, **kwargs):
        Voronoi.__init__(self, scaler)
        self.tilt = tilt
        self.min_probability = min_probability
        self.avg_probability = avg_probability
        self.max_probability = max_probability
        self.min_distance = min_distance
        self.avg_distance = avg_distance

    def tessellate(self, points, **kwargs):
        if isinstance(points, pd.DataFrame):
            points = points.copy()
        points = self._preprocess(points)
        # center
        mean_kwargs = {}
        if not isinstance(points, pd.DataFrame):
            mean_kwargs['keepdims'] = True
        center = self.descriptors(points).mean(axis=0, **mean_kwargs)
        points -= center
        if isinstance(points, pd.DataFrame):
            _center = center[self.scaler.columns]
            center = center.values
            pts = points[self.scaler.columns].values
        else:
            pts = points
        # bounds
        lower_bound = kwargs.get('lower_bound', None)
        if lower_bound is None:
            lower_bound = points.min(axis=0)
        elif isinstance(points, pd.DataFrame):
            lower_bound -= _center
        else:
            lower_bound -= center
        upper_bound = kwargs.get('upper_bound', None)
        if upper_bound is None:
            upper_bound = points.max(axis=0)
        elif isinstance(points, pd.DataFrame):
            upper_bound -= _center
        else:
            upper_bound -= center
        # rotate
        if self.tilt:
            theta = self.tilt * pi / 6.
            s, c = sin(-theta), cos(-theta)
            rot = np.array([[c, -s], [s, c]])
            if isinstance(points, pd.DataFrame):
                pts = np.dot(pts, rot.T)
                points.loc[:,self.scaler.columns] = pts
            else:
                points = pts = np.dot(points, rot.T)
        #
        x0, y0 = (lower_bound + upper_bound) * .5
        size = upper_bound - lower_bound
        hex_sin, hex_cos = sin(pi/6.), cos(pi/6.)
        if self.avg_probability:
            desired_n_cells = n_cells = 1. / self.avg_probability
            def err(r):
                if r == 0:
                    return desired_n_cells * desired_n_cells
                dx, dy = 2 * r, 1.5 * r / hex_cos
                m, n = ceil(size[0] / dx), ceil(size[1] / dy) + 1
                e = m * n + ceil(n / 2) - desired_n_cells
                return e * e + r * r
            hex_radius = minimize_scalar(err).x
            if self.min_distance is not None:
                hex_radius = max(hex_radius, .5 * self.min_distance)
            hex_side = hex_radius / hex_cos
            dx = 2. * hex_radius
            dy = 1.5 * hex_side# * (2. - hex_sin)
            m = int(ceil(size[0] / dx))
            n = int(ceil((size[1] + 2. * (hex_side * hex_sin - hex_radius)) / dy) + 1)
            lower_center_x = x0 -.5 * float(m) * dx
            lower_center_y = y0 -.5 * float(n - 1) * dy
            centers = []
            for k in range(n):
                _centers_x = lower_center_x + float(k % 2) * hex_radius + dx * np.arange(m + 1 - (k % 2))
                _centers_y = np.full_like(_centers_x, lower_center_y + float(k) * dy)
                centers.append(np.stack((_centers_x, _centers_y), axis=-1))
            self._cell_centers = np.vstack(centers)
            I = np.unique(self.cell_index(points))
            assert np.all(0 <= I)
            n_cells = I.size
        elif self.avg_distance:
            hex_radius = .5 * self.avg_distance
            if self.min_probability is not None:
                # TODO
                pass
            hex_side = hex_radius / hex_cos
            dx = 2. * hex_radius
            dy = 1.5 * hex_side# * (2. - hex_sin)
            m = int(ceil(size[0] / dx))
            n = int(ceil((size[1] + 2. * (hex_side * hex_sin - hex_radius)) / dy) + 1)
            lower_center_x = x0 -.5 * float(m) * dx
            lower_center_y = y0 -.5 * float(n - 1) * dy
            centers = []
            for k in range(n):
                _centers_x = lower_center_x + float(k % 2) * hex_radius + dx * np.arange(m + 1 - (k % 2))
                _centers_y = np.full_like(_centers_x, lower_center_y + float(k) * dy)
                centers.append(np.stack((_centers_x, _centers_y), axis=-1))
            self._cell_centers = np.vstack(centers)
        else:
            raise ValueError('both `avg_probability` and `avg_distance` undefined')
        # rotate/center the centers back into the original space
        if self.tilt:
            s, c = sin(theta), cos(theta)
            rot = np.array([[c, -s], [s, c]])
            self._cell_centers = np.dot(self._cell_centers, rot.T)
        self._cell_centers += center
        self.hexagon_radius = hex_radius
        self.hexagon_count = (m, n)

    def _preprocess(self, points):
        if points.shape[1] != 2:
            msg = 'the number of dimensions is not 2'
            try:
                points = points[['x', 'y']]
            except:
                raise ValueError(msg)
            else:
                warn(msg, RuntimeWarning)
        Voronoi._preprocess(self, points) # initialize `scaler`
        return points # ... but do not scale

    # cell_centers property
    @property
    def cell_centers(self):
        return self._cell_centers

    @cell_centers.setter
    def cell_centers(self, centers):
        self._cell_centers = centers

    # cell_adjacency property
    @property
    def cell_adjacency(self):
        if self._cell_adjacency is None:
            n_cells = self._cell_centers.shape[0]
            m, n = self.hexagon_count
            A = sparse.dok_matrix((n_cells, n_cells), dtype=bool)
            k = i = 0
            for k in range(n-1):
                m_no_more = k % 2
                A[i, i+1] = True
                A[i, i+m+1] = True
                if m_no_more:
                    A[i, i+m] = True
                i += 1
                for j in range(1, m - m_no_more):
                    A[i, i+1] = True
                    A[i, i+m+1] = True
                    A[i, i+m] = True
                    i += 1
                if m_no_more:
                    A[i, i+m+1] = True
                A[i, i+m] = True
                i += 1
            k += 1
            for j in range(m - (k % 2)):
                A[i, i+1] = True
                i += 1
            assert A.shape == (n_cells, n_cells)
            self._cell_adjacency = (A + A.T).tocsr()
        return self.__returnlazy__('cell_adjacency', self._cell_adjacency)

    @cell_adjacency.setter # copy/paste
    def cell_adjacency(self, matrix):
        self.__lazysetter__(matrix)

    def _postprocess(self):
        if self._cell_centers is None:
            raise NameError('`cell_centers` not defined; tessellation has not been grown yet')
        s, c = sin(pi/6.), cos(pi/6.)
        dr = np.array([[0., -1.], [-c, -s], [-c, s], [c, -s], [c, s], [0., 1.]])
        if self.tilt:
            theta = self.tilt * pi / 6.
            _s, _c = sin(theta), cos(theta)
            rot = np.array([[_c, -_s], [_s, _c]])
            dr = np.dot(dr, rot.T)
        dr *= self.hexagon_radius / c
        #
        n_cells = self._cell_centers.shape[0]
        m, n = self.hexagon_count
        C = self._cell_centers
        _v = np.full((n_cells, 6), -1, dtype=int)
        # bottom left hexagon
        i = 0
        V = [C[i]+dr] # full hexagon
        _v[i] = np.arange(6)
        A = [(0,1), (0,3), (1,2), (2,5), (3,4), (4,5)]
        v = 6
        # vertices are numbered the following way:
        #
        #     5
        #  2     4
        #  1     3
        #     0
        #
        for j in range(1, m+1):
            i += 1
            V.append(C[i]+dr[[0,3,4,5]])
            _v[i, 1] = v - 3
            _v[i, 2] = v - 2
            _v[i, [0,3,4,5]] = np.arange(v, v+4)
            A += [(v-3,v), (v-2,v+3), (v,v+1), (v+1,v+2), (v+2,v+3)]
            v += 4
        i_prev = -1 # index of the bottom right neighbour (should be 0 after first increment)
        for k in range(1, n):
            m_plus = 1 - (k % 2)
            i += 1
            i_prev += 1
            # i_prev is the first (left-most) cell of the previous row
            if m_plus:
                # i_prev is already the bottom right neighbour
                V.append(C[i]+dr[[1,2,4,5]])
                _v[i, 0] = _v[i_prev, 2]
                _v[i, 3] = _v[i_prev, 5]
                _v[i, [1,2,4,5]] = np.arange(v, v+4)
                A += [
                    # bottom right 2 (=0) -> 1
                    (_v[i_prev,2],v),
                    # bottom right 5 (=3) -> 4
                    (_v[i_prev,5],v+2),
                    # 1 -> 2
                    (v,v+1),
                    # 2 -> 5
                    (v+1,v+3),
                    # 4 -> 5
                    (v+2,v+3),
                    ]
                v += 4
            else:
                # i_prev points to the bottom left neighbour
                i_prev += 1 # now bottom right
                V.append(C[i]+dr[[2,4,5]])
                _v[i, 0] = _v[i_prev, 2]
                _v[i, 1] = _v[i_prev-1, 5]
                _v[i, 3] = _v[i_prev, 5]
                _v[i, [2,4,5]] = np.arange(v, v+3)
                A += [
                    # bottom left 5 (=1) -> 2
                    (_v[i_prev-1,5],v),
                    # bottom right 5 (=3) -> 4
                    (_v[i_prev,5],v+1),
                    # 2 -> 5
                    (v,v+2),
                    # 4 -> 5
                    (v+1,v+2),
                    ]
                v += 3
            for j in range(1, m):
                i += 1
                i_prev += 1
                # j stops at m => there is always a bottom right neighbour => vertex 3 already exists
                V.append(C[i]+dr[[4,5]])
                _v[i, 0] = _v[i_prev, 2]
                _v[i, 1] = _v[i-1, 3] # =v-3
                _v[i, 2] = _v[i-1, 4] # =v-2
                _v[i, 3] = _v[i_prev, 5]
                _v[i, [4,5]] = np.arange(v, v+2)
                A += [
                    # bottom right 5 (=3) -> 4
                    (_v[i_prev,5],v),
                    # left 4 (=2) -> 5
                    (v-2,v+1),
                    # 4 -> 5
                    (v,v+1),
                    ]
                v += 2
            if m_plus:
                i += 1
                # there is no bottom right neighbour
                # let i_prev point to the bottom left neighbour
                V.append(C[i]+dr[[3,4,5]])
                _v[i, 0] = _v[i_prev, 4]
                _v[i, 1] = _v[i-1, 3] # =v-3
                _v[i, 2] = _v[i-1, 4] # =v-2
                _v[i, [3,4,5]] = np.arange(v, v+3)
                A += [
                    # bottom left 4 (=0) -> 3
                    (_v[i_prev,4],v),
                    # left 4 (=2) -> 5
                    (v-2,v+2),
                    # 3 -> 4
                    (v,v+1),
                    # 4 -> 5
                    (v+1,v+2),
                    ]
                v += 3
        self._vertices = np.vstack(V)
        n_vertices = self._vertices.shape[0]
        i, j = zip(*A)
        i, j = np.array(i+j), np.array(j+i)
        self._vertex_adjacency = sparse.csr_matrix(
            (np.ones(i.size, dtype=bool), (i, j)),
            shape=(n_vertices, n_vertices))
        self._cell_vertices = { i: v[0<=v] for i, v in enumerate(_v) }

    # vertices property
    @property
    def vertices(self):
        if self._vertices is None and self._cell_centers is not None:
            self._postprocess()
        return self.__returnlazy__('vertices', self._vertices) # no scaling

    @vertices.setter # copy/paste
    def vertices(self, vertices):
        self.__lazysetter__(vertices)

    # cell_volume property
    @property
    def cell_volume(self):
        if self._cell_volume is None:
            n_cells = self._cell_centers.shape[0]
            cell_area = 3 * self.hexagon_radius ** 2 / cos(pi/6.)
            self._cell_volume = np.full(n_cells, cell_area, dtype=float)
        return self._cell_volume

    @cell_volume.setter
    def cell_volume(self, area):
        self.__setlazy__('cell_volume', area)




setup = {
    'make_arguments': OrderedDict((
        ('tilt', dict(type=float, default=0., help='angle of a main axis, in pi/6 radians; from: 0= a main axis is the x-axis; to: 1= a main axis is the y-axis')),
        ('min_probability', ()),
        ('avg_probability', ()),
        ('max_probability', ()),
        ('min_distance', ()),
        ('avg_distance', ()),
        # avg_location_count allows to control avg_probability from the commandline
        ('avg_location_count', dict(args=('-c', '--location-count'), kwargs=dict(type=int, default=80, help='average number of locations per cell'), translate=True)),
        )),
    }


__all__ = ['setup', 'HexagonalMesh']

import sys
if sys.version_info[0] < 3:

    import rwa
    import tramway.core.hdf5 as rules
    hexagonal_mesh_exposes = rules.voronoi_exposes + list(HexagonalMesh.__slots__)
    rwa.hdf5_storable(
        rwa.default_storable(HexagonalMesh, exposes=hexagonal_mesh_exposes),
        agnostic=True)

    __all__.append('hexagonal_mesh_exposes')




# -*- coding: utf-8 -*-

# Copyright © 2017-2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


import os
import numpy as np
import pandas as pd
import scipy.sparse as sparse
from ..core import *
from ..core.hdf5 import *
from ..core.analyses import abc
from ..tessellation import *
from .base import *
from warnings import warn
import six
import traceback
import itertools
# no module-wide matplotlib import for head-less usage of `tessellate`
# in the case matplotlib's backend is interactive


class Tessellate(Helper):
    def __init__(self):
        Helper.__init__(self)
        self._label_is_output = True
        self.plugins = plugins
        self.tessellation_kwargs = {}
        self.partition_kwargs = {}
        self.time_window_kwargs = {}
        self.reassignment_kwargs = {}

    @property
    def _partition_kwargs(self):
        if self.reassignment_kwargs:
           # reassigning is based on true partitions (Voronoi),
           # therefore over- and under-sampling arguments should be silenced
           over_sampling_args = ('knn', 'radius', 'time_knn')
           return { kw: arg for kw, arg in self.partition_kwargs.items() \
                        if kw not in over_sampling_args }
        else:
           return self.partition_kwargs

    @_partition_kwargs.setter
    def _partition_kwargs(self, kwargs):
        self.partition_kwargs = kwargs

    def prepare_data(self, input_data, labels=None, types=None, metadata=True, \
            verbose=None, scaling=False, time_scale=None, **kwargs):

        if (isinstance(input_data, six.string_types) and os.path.isdir(input_data)) or \
                (self.are_multiple_files(input_data) and input_data and all(os.path.exists(_d) for _d in input_data)):
            input_data, self.input_file = load_xyt(input_data, return_paths=True, **kwargs)
            self.metadata['datafile'] = self.input_file if self.input_file[1:] else self.input_file[0]
        elif isinstance(input_data, six.string_types):
            if not os.path.exists(input_data):
                raise OSError('file not found: {}'.format(input_data))
            try:
                input_data, self.input_file = load_xyt(input_data, return_paths=True, **kwargs)
            except (KeyboardInterrupt, SystemExit):
                raise
            except (UnicodeDecodeError, pd.errors.ParserError): # rwa file
                pass
            else:
                assert self.input_file and not self.input_file[1:]
                self.metadata['datafile']= self.input_file[0]

        if labels is None:
            labels = self.input_label
        nesting = labels is not None
        if types is None and nesting:
            types = Partition

        data = Helper.prepare_data(self, input_data, labels, types, metadata, verbose, **kwargs)

        if nesting:
            assert isinstance(data, tuple) and not data[1:]
            data, = data
            if isinstance(data, Partition):
                self.input_partition = data
                self.xyt_data = data.points
            else:
                if isinstance(data, pd.DataFrame):
                    msg = '(trans-)location data'
                else:
                    msg = 'the following datatype: {}'.format(type(data))
                raise TypeError('nesting tessellations does not apply to {}'.format(msg))
            if self.input_label is None:
                self.input_label = self.find(data)
        elif isinstance(data, abc.Analyses):
            self.xyt_data = data.data
        else:
            self.xyt_data = data

        if scaling:
            if scaling is True:
                scaling = 'whiten'
            self.scaler = dict(whiten=whiten, unit=unitrange)[scaling]()
        else:
            self.scaler = None

        self.colnames = ['x', 'y']
        if 'z' in self.xyt_data:
            self.colnames.append('z')
        if time_scale:
            self.colnames.append('t')
            self.scaler.factor = [('t', time_scale)]

        return data

    def output_file(self, output_file=None, suffix=None, extension=None):
        none_defined = output_file is None and suffix is None and extension is None
        output_file = Helper.output_file(self, output_file, suffix, extension)
        if none_defined and self.input_file:
            output_file = os.path.splitext(output_file)[0] + '.rwa'
        return output_file

    def plugin(self, method, plugins=None, verbose=None):
        Helper.plugin(self, method, plugins, verbose)
        self.constructor = self.setup['make']
        if isinstance(self.constructor, str):
            self.constructor = getattr(self.module, self.setup['make'])

    def standard_parameters(self, \
            distance=None, ref_distance=None, \
            rel_min_distance=None, rel_avg_distance=None, rel_max_distance=None, \
            min_location_count=None, avg_location_count=None, max_location_count=None, \
            knn=None, radius=None, time_knn=None, verbose=None):
        if verbose is None:
            verbose = self.verbose
        transloc_length = min_distance = avg_distance = max_distance = None
        if ref_distance is None and distance is not None:
            # `distance` is only for compatibility with the tramway commandline
            ref_distance = distance
        nesting = self.input_label is not None
        if nesting and self.input_partition.param is not None:
            prm = self.input_partition.param.get('tessellation', self.input_partition.param)
            if not ref_distance:
                ref_distance = prm.get('ref_distance', None)
            min_distance = prm.get('min_distance', min_distance)
            avg_distance = prm.get('avg_distance', avg_distance)
            max_distance = prm.get('max_distance', max_distance)
        # former default values for `rel_min_distance` and `rel_avg_distance`
        if rel_min_distance is None and min_distance is None:
            rel_min_distance = .8
        if rel_avg_distance is None and avg_distance is None:
            rel_avg_distance = 2.

        if ref_distance is None:
            if 'n' not in self.xyt_data.columns:
                raise ValueError('please specify ref_distance')
            transloc_xy = np.asarray(translocations(self.xyt_data))
            if transloc_xy.shape[0] == 0:
                raise ValueError('no translocation found')
            transloc_length = np.nanmean(np.sqrt(np.sum(transloc_xy * transloc_xy, axis=1)))
            if verbose:
                print('average translocation distance: {}'.format(transloc_length))
            ref_distance = transloc_length
        if rel_min_distance is not None:
            min_distance = rel_min_distance * ref_distance
        if rel_avg_distance is not None:
            avg_distance = rel_avg_distance * ref_distance
        if rel_max_distance is not None:
            # applies only to KDTreeMesh and Kohonen
            max_distance = rel_max_distance * ref_distance

        if min_location_count is None: # former default value: 20
            if knn is None and radius is None:
                min_location_count = 20
        n_pts = float(self.xyt_data.shape[0])
        if min_location_count:
            min_probability = float(min_location_count) / n_pts
        else:
            min_probability = None
        if avg_location_count is None:
            if min_location_count is None:
                avg_location_count = 80 # former default value
            else:
                avg_location_count = 4 * min_location_count
        if avg_location_count:
            avg_probability = float(avg_location_count) / n_pts
        else:
            avg_probability = None
        if max_location_count:
            max_probability = float(max_location_count) / n_pts
        else:
            max_probability = None

        params = dict( \
            ref_distance=ref_distance, \
            min_distance=min_distance, \
            avg_distance=avg_distance, \
            max_distance=max_distance, \
            min_probability=min_probability, \
            avg_probability=avg_probability, \
            max_probability=max_probability, \
            min_location_count=min_location_count, \
            avg_location_count=avg_location_count, \
            max_location_count=max_location_count, \
            )

        return params


    def parse_args(self, params, knn=None, radius=None,
            rel_max_size=None, rel_max_volume=None, \
            time_window_duration=None, time_window_shift=None, \
            enable_time_regularization=False, time_window_options=None, \
            min_n=None, time_knn=None, kwargs={}):
        for ignored in ['max_level']:
            try:
                if self.tessellation_kwargs[ignored] is None:
                    del self.tessellation_kwargs[ignored]
            except KeyError:
                pass
        params = dict(params)
        params.update(self.tessellation_kwargs)
        for key in self.setup.get('make_arguments', {}):
            try:
                param = params[key]
            except KeyError:
                pass
            else:
                self.tessellation_kwargs[key] = param

        self.time_window_kwargs = {}
        if time_window_duration:
            self.time_window_kwargs['duration'] = time_window_duration
        if time_window_shift:
            self.time_window_kwargs['shift'] = time_window_shift
        if time_window_options:
            time_window_options = dict(time_window_options) # copy to prevent side effects
            enable_time_regularization = time_window_options.pop(
                        'regularisable',
                        time_window_options.pop(
                            'regularizable',
                            enable_time_regularization))
            self.time_window_kwargs.update(time_window_options)
        if enable_time_regularization:
            time_dimension = self.time_window_kwargs.get('time_dimension', None)
            if time_dimension is False:
                warn('`enable_time_regularization={}` and `time_dimension={}` are not compatible'.format(enable_time_regularization, time_dimension), RuntimeWarning)
            elif time_dimension is None:
                self.time_window_kwargs['time_dimension'] = True
        if self.time_window_kwargs:
            self.time_window_kwargs['duration'] # required; fails with KeyError if missing

        ref_distance = params.get('ref_distance', None)
        _filter_f = self.partition_kwargs.get('filter', None)
        try:
            max_size = kwargs['rel_max_size']
        except KeyError:
            _filter_fg = _filter_f
        else:
            max_size *= ref_distance
            def _filter_g(voronoi, cell, points):
                return voronoi.scaler.unscale_distance(np.sqrt(np.min(np.sum(( \
                        voronoi._center[cell] - \
                        voronoi._vertices[voronoi.cell_vertices[cell]] \
                    )*2, axis=1)))) <= max_size
            _filter_fg = _filter_g if _filter_f is None \
                else lambda *a: _filter_g(*a) and _filter_f(*a)
        try:
            max_volume = kwargs['rel_max_volume']
        except KeyError:
            _filter_fgh = _filter_fg
        else:
            max_volume *= ref_distance
            def _filter_h(voronoi, cell, points):
                return voronoi.cell_volume[cell] <= max_volume
            _filter_fgh = _filter_h if _filter_fg is None \
                else lambda *a: _filter_h(*a) and _filter_fg(*a)
        try:
            max_actual_size = self.partition_kwargs.pop('rel_max_size')
        except KeyError:
            _filter_fghi = _filter_fgh
        else:
            max_actual_size *= ref_distance
            _descr = not self.partition_kwargs.get('filter_descriptors_only', True)
            def _filter_i(voronoi, cell, x):
                if _descr:
                    x = voronoi.descriptors(x, asarray=True)
                x2 = np.sum(x * x, axis=1, keepdims=True)
                d2 = x2 + x2.T - 2. * np.dot(x, x.T)
                d = np.sqrt(np.max(d2.ravel()))
                return d <= max_actual_size
            _filter_fghi = _filter_i if _filter_fgh is None \
                else lambda *a: _filter_i(*a) and _filter_fgh(*a)
        if _filter_fghi is not None:
            self.partition_kwargs['filter'] = _filter_fghi
            if 'filter_descriptors_only' not in self.partition_kwargs:
                self.partition_kwargs['filter_descriptors_only'] = True
        if not (knn is None and radius is None and time_knn is None):
            if knn is not None:
                self.partition_kwargs['knn'] = knn
            if radius is not None:
                if 'radius' in self.partition_kwargs:
                    warn('overwriting `radius`', RuntimeWarning)
                self.partition_kwargs['radius'] = radius
            if time_knn is not None:
                self.partition_kwargs['time_knn'] = time_knn
            if 'min_location_count' not in self.partition_kwargs:
                min_location_count = params['min_location_count']
                self.partition_kwargs['min_location_count'] = min_location_count
            if 'metric' not in self.partition_kwargs:
                self.partition_kwargs['metric'] = 'euclidean'
        if min_n is not None:
            self.partition_kwargs['min_location_count'] = min_n

        _options = kwargs.pop('reassignment_options', {})
        if _options:
            self.reassignment_kwargs.update(_options)
        for arg in dict(kwargs):
            if arg.startswith('reassignment_'):
                self.reassignment_kwargs[arg[13:]] = kwargs.pop(arg)
            if arg.startswith('reassign_'):
                self.reassignment_kwargs[arg[9:]] = kwargs.pop(arg)

    def tessellate(self, comment=None, verbose=None):
        if comment is None:
            comment = self.comment
            if self.comment is not None:
                self.comment = None
        if verbose is None:
            verbose = self.verbose

        nesting = self.input_label is not None

        # initialize a Tessellation object
        if nesting:
            if self.time_window_kwargs:
                raise NotImplementedError('spatial tessellation combined with time windowing is not supported yet by tessellation nesting')
            tess = NestedTessellations(self.scaler, self.input_partition, factory=self.constructor,
                **self.tessellation_kwargs)
        else:
            if self.time_window_kwargs:# and not self.reassignment_kwargs:
                # note: bin reassignment applies only to the spatial component;
                #       time windowing is carried out after reassigning bins
                import tramway.tessellation.window as window
                tess = window.SlidingWindow(**self.time_window_kwargs)
                tess.spatial_mesh = self.constructor(self.scaler, **self.tessellation_kwargs)
            else:
                tess = self.constructor(self.scaler, **self.tessellation_kwargs)

        # grow the tessellation
        if nesting:
            data = self.xyt_data
        else:
            data = self.xyt_data[self.colnames]
        tessellate_kwargs = self.tessellation_kwargs
        tessellate_hidden_kwargs = {}
        if verbose is not None:
            tessellate_hidden_kwargs['verbose'] = verbose
        tess.tessellate(data, **tessellate_kwargs, **tessellate_hidden_kwargs)

        # partition the dataset into the cells of the tessellation
        try:
            cell_index = tess.cell_index(self.xyt_data, **self._partition_kwargs)
        except MemoryError:
            if verbose:
                print(traceback.format_exc())
            warn('memory error: cannot assign points to cells', RuntimeWarning)
            cell_index = None

        self.cells = cells = Partition(self.xyt_data, tess, cell_index)

        if self.reassignment_kwargs:
            while True:
                ncells = cells.number_of_cells
                cells = self.reassign()
                if cells.number_of_cells == ncells:
                    break
                #assert self.cells.cell_index.max() < cells.number_of_cells
            if cells.number_of_cells < 10:
                warn('coarse tessellation: {} cells only'.format(cells.number_of_cells), RuntimeWarning)
            # post-reassignment step to introduce overlap and time windowing if requested
            if self.time_window_kwargs:
                cells = self.cells
                tess = cells.tessellation
            if self._partition_kwargs is not self.partition_kwargs:
                cell_index = tess.cell_index(self.xyt_data, **self.partition_kwargs)
                if isinstance(cell_index, tuple) and len(cell_index[0]) == 0:
                    print('time_window_kwargs', self.time_window_kwargs)
                    print('partition_kwargs', self.partition_kwargs)
                    print('cell_index', cell_index)
                    print('data.shape', self.xyt_data.shape, 'number_of_cells', tess.number_of_cells)
                    raise ValueError('not any point assigned')
                #self.cells = cells = Partition(self.xyt_data, tess, cell_index)
                self.cells.cell_index = cell_index

        # store some parameters together with the partition
        method = self.name
        cells.param['method'] = method
        if self.time_window_kwargs:
            cells.param['time_window'] = self.time_window_kwargs
        if self.tessellation_kwargs:
            cells.param['tessellation'] = self.tessellation_kwargs
        if self.partition_kwargs:
            cells.param['partition'] = self.partition_kwargs
        if self.reassignment_kwargs:
            cells.param['reassignment'] = self.reassignment_kwargs

        # insert the resulting analysis in the analysis tree
        if self.analyses is not None:
            self.insert_analysis(cells, comment=comment)

        return cells

    def reassign(self):
        """called by :met:`tessellate`. Should not be called directly."""
        cells = self.cells

        #partition_kwargs = dict(cells.param.get('partition',{}))
        #partition_kwargs.update(self.partition_kwargs)
        partition_kwargs = self._partition_kwargs

        if self.time_window_kwargs:
            import copy
            tess = cells.tessellation.spatial_mesh
            point_index,cell_index = cells.cell_index
            cell_index = cell_index % tess.number_of_cells
            def noway(*args, **kwargs):
                raise RuntimeError('cell overlap is not supported in combination with point reassignment')
            cell_index = format_cell_index((point_index,cell_index),
                    format='array', select=noway, shape=(len(cells.points),))
            cells = Partition(cells.points, tess, cell_index)

        update_centroids = self.reassignment_kwargs.get('update_centroids', False)

        reassignment_count_threshold = self.reassignment_kwargs['count_threshold']
        reassignment_priority = self.reassignment_kwargs.get('priority_by', 'count')

        label = True
        while True:
            if update_centroids:
                if cells._cell_index is None:
                    warn('updating the centroids after reassigning points may fail because of memory error', RuntimeWarning)
                prev_cell_indices = cells.cell_index
                if not isinstance(prev_cell_indices, np.ndarray):
                    raise RuntimeError('cell overlap is not supported in combination with point reassignment')

            cells, deleted_cells, label = delete_low_count_cells(cells,
                    reassignment_count_threshold, reassignment_priority, label, self._partition_kwargs)

            if deleted_cells.size == 0:
                break

            dim = cells.tessellation.cell_centers.shape[1]
            min_ncells = dim + 2
            if cells.number_of_cells < min_ncells:
                raise RuntimeError('too few remaining cells ({}<{})'.format(cells.number_of_cells, min_ncells))

            if update_centroids:
                cells = update_cell_centers(cells, update_centroids, partition_kwargs)
            if self.time_window_kwargs:
                self.cells.tessellation.spatial_mesh = tess = cells.tessellation
                self.cells.cell_index = point_index,cell_index = \
                        self.cells.tessellation.cell_index(cells.points, **partition_kwargs)
                cell_index = cell_index % tess.number_of_cells
                cells.cell_index = cell_index = \
                        format_cell_index((point_index,cell_index),
                                format='array', select=noway, shape=(len(cells.points),))

        if not self.time_window_kwargs:
            self.cells = cells
        return cells


def tessellate1(xyt_data, method='gwr', output_file=None, verbose=False, \
        scaling=False, time_scale=None, \
        knn=None, radius=None, distance=None, ref_distance=None, \
        rel_min_distance=None, rel_avg_distance=None, rel_max_distance=None, \
        min_location_count=None, avg_location_count=None, max_location_count=None, \
        rel_max_size=None, rel_max_volume=None, \
        time_window_duration=None, time_window_shift=None, time_window_options=None, \
        enable_time_regularization=False, time_knn=None, \
        label=None, output_label=None, comment=None, input_label=None, inplace=False, \
        overwrite=None, return_analyses=False, \
        load_options=None, tessellation_options=None, partition_options=None, save_options=None, \
        force=None, \
        **kwargs):
    """
    Tessellation from points series and partitioning.

    This helper routine is a high-level interface to the various tessellation techniques
    implemented in TRamWAy.

    In addition to `knn`, `radius`, *filter* and *metric*, arguments with prefix *strict_* in
    their name apply to the partitioning step only, while the others apply to the tessellation step.

    *rel_max_size* and *rel_max_volume* are notable exceptions in that they currently apply to
    the partitioning step whereas they should conceptually apply to the tessellation step instead.
    This may change in a future version.

    Reassignment options can be provided using the `reassignment_options` dictionnary or keyworded
    arguments which names begin with '*reassign_*' or '*reassignment_*'.

    Arguments:
        xyt_data (str or pandas.DataFrame):
            Path to a *.trxyt* or *.rwa* file or raw data in the shape of
            :class:`pandas.DataFrame`.


        method (str):
            Tessellation method or plugin name.
            See for example
            :class:`~tramway.tessellation.random.RandomMesh` ('*random*'),
            :class:`~tramway.tessellation.grid.RegularMesh` ('*grid*'),
            :class:`~tramway.tessellation.hexagon.HexagonalMesh` ('*hexagon*'),
            :class:`~tramway.tessellation.kdtree.KDTreeMesh` ('*kdtree*'),
            :class:`~tramway.tessellation.kmeans.KMeansMesh` ('*kmeans*') and
            :class:`~tramway.tessellation.gwr.GasMesh` ('*gas*' or '*gwr*').

        output_file (str):
            Path to a *.rwa* file. The resulting tessellation and data partition will be
            stored in this file. If `xyt_data` is a path to a file and `output_file` is not
            defined, then `output_file` will be adapted from `xyt_data` with extension
            *.rwa* and possibly overwrite the input file.

        verbose (bool or int): Verbose output.

        scaling (bool or str):
            Normalization of the data.
            Any of '*unitrange*', '*whiten*' or other methods defined in
            :mod:`tramway.core.scaler`.

        time_scale (bool or float):
            If this argument is defined and intepretable as ``True``, the time axis is
            scaled by this factor and used as a space variable for the tessellation (2D+T or
            3D+T, for example).
            This is equivalent to manually scaling the ``t`` column and passing
            ``scaling=True``.

        knn (int or pair of ints):
            After growing the tessellation, a minimum and maximum numbers of nearest
            neighbours of each cell center can be used instead of the entire cell
            population. Let us denote ``min_nn, max_nn = knn``. Any of ``min_nn`` and
            ``max_nn`` can be ``None``.
            If a single `int` is supplied instead of a pair, then `knn` becomes ``min_nn``.
            ``min_nn`` enables cell overlap and any point may be associated with several
            cells.
            See also :meth:`~tramway.tessellation.base.Delaunay.cell_index`.

        radius (float):
            After growing the tessellation as a set of centroids, a cell will consist of
            the locations within this distance from the centroid.
            See also :meth:`~tramway.tessellation.base.Delaunay.cell_index`.

        distance/ref_distance (float):
            Supposed to be the average translocation distance. Can be modified so that the
            cells are smaller or larger.

        rel_min_distance (float):
            Multiplies with `ref_distance` to define the minimum inter-cell distance.

        rel_avg_distance (float):
            Multiplies with `ref_distance` to define an upper on the average inter-cell
            distance.

        rel_max_distance (float):
            Multiplies with `ref_distance` to define the maximum inter-cell distance.

        min_location_count (int):
            Minimum number of points per cell. Depending on the method, can be strictly
            enforced or regarded as a recommendation.

        avg_location_count (int):
            Average number of points per cell. For non-plugin method, per default, it is
            set to four times `min_location_count`.

        max_location_count (int):
            Maximum number of points per cell. This is used by *kdtree* and *gwr*.

        rel_max_size (float):
            Maximum cell diameter as a number of `ref_distance`. Diameter (or size) is
            estimated as twice the distance between the center of cell and the nearest
            vertex. Cells of excess size are ignored so as the associated locations.

        rel_max_volume (float):
            Maximum cell volume (or surface area in 2D) as a number of `ref_distance`.
            Cells of excess volume are ignored so as the associated locations.

        strict_min_location_count/min_n (int):
            Minimum number of points per cell in the eventual partition. Cells with
            insufficient points are ignored so as the associated locations.

        strict_rel_max_size (float):
            Maximum cell diameter as a number of `ref_distance`. Diameter (or size) is
            estimated as the maximum distance between any pair of locations in the cell.
            Cells of excess size are ignored so as the associated locations.

        time_window_duration (float):
            Window duration in seconds (or frames with
            ``time_window_options=dict(frames=True)``).
            This time windowing combines with any other spatial tessellation method.
            To use the :mod:`~tramway.tessellation.window` plugin only, use ``method=window``
            and its *duration* and *shift* arguments instead.
            See also the :mod:`~tramway.tessellation.window` plugin.

        time_window_shift (float):
            Window shift in seconds (or frames with
            ``time_window_options=dict(frames=True)``).
            Default is no overlap, i.e. ``time_window_shift=time_window_duration``.
            See also the :mod:`~tramway.tessellation.window` plugin.

        time_window_options (dict):
            Extra arguments for time windowing.
            See also the :mod:`~tramway.tessellation.window` plugin.

        enable_time_regularization (bool):
            Equivalent to ``time_window_options['time_dimension'] = enable_time_regularization``.

        time_knn (int or pair of ints):
            ``min_nn`` minimum number of nearest "neighbours" of the time segment center,
            or ``(min_nn, max_nn)`` minimum and maximum numbers of nearest neighbours in time.
            See also :meth:`~tramway.tessellation.time.TimeLattice.cell_index`.

        input_label (str):
            Label for the input tessellation for nesting tessellations.

        label/output_label (int or str):
            Label for the resulting analysis instance.

        inplace (bool):
            If True, `label`/`output_label`/`input_label` are exclusive, they all define
            a same analysis and the resulting analysis replaces the input analysis.

        comment (str):
            Description message for the resulting analysis.

        overwrite (bool): if an implicit output file already exists, overwrite it.

        return_analyses (bool):
            Return a :class:`~tramway.core.analyses.base.Analyses` object instead of
            the default :class:`~tramway.tessellation.base.Partition` output.

        load_options (dict):
            Pass extra keyword arguments to :func:`~tramway.core.xyt.load_xyt` if called.

        tessellation_options (dict):
            Pass explicit keyword arguments to the *__init__* function of the
            tessellation class as well as to the
            :meth:`~tramway.tessellation.base.Tessellation.tessellate` method, and ignore
            the extra input arguments.

        partition_options (dict):
            Pass explicit keyword arguments to the
            :meth:`~tramway.tessellation.base.Tessellation.cell_index` method and ignore
            the extra input arguments.

        reassignment_options (dict):
            allowed keys are:
            *count_threshold* (int): minimum number of points for a cell not to be deleted;
            *update_centroids* (bool or int): if evaluates to ``True``, cell centers are
            updated as the centers of mass for the assigned points, and then the new cells
            are checked again for *count_threshold* and so on; the number of iterations is
            not limited by default, and can be set passing an integer value instead of ``True``.

        save_options (dict):
            Pass extra keyword arguments to :func:`~tramway.core.xyt.save_rwa` if called.

    Returns:
        tramway.tessellation.base.Partition: A partition of the data with its
            :attr:`~tramway.tessellation.base.Partition.tessellation` attribute set.


    Apart from the parameters defined above, extra input arguments are admitted and may be passed
    to the initializer of the selected tessellation method as well as to the
    :meth:`~tramway.tessellation.base.Tessellation.tessellate` and
    :meth:`~tramway.tessellation.base.Tessellation.cell_index` methods.

    See the individual documentation of these methods for more information.

    """
    helper = Tessellate()
    helper.verbose = verbose
    helper.labels(label=label, input_label=input_label, output_label=output_label, inplace=inplace)
    if load_options is None:
        load_options = {}
    if helper.are_multiple_files(xyt_data) and len(xyt_data) == 1:
        xyt_data = next(iter(xyt_data))
    helper.prepare_data(xyt_data, scaling=scaling, time_scale=time_scale,
            metadata=not kwargs.pop('disable_metadata',None), **load_options)
    helper.plugin(method)

    # distinguish between tessellation and partition arguments
    if tessellation_options is None and partition_options is None:
        helper.tessellation_kwargs = dict(kwargs)
        helper.partition_kwargs = {}
        for _kw in kwargs:
            if _kw.startswith('strict_'):
                helper.partition_kwargs[_kw[7:]] = helper.tessellation_kwargs.pop(_kw)
        for _kw in ('filter', 'filter_descriptors_only', 'metric'):
            try:
                _arg = helper.tessellation_kwargs.pop(_kw)
            except KeyError:
                pass
            else:
                helper.partition_kwargs[_kw] = _arg
        for _kw in ('rel_max_size', 'rel_max_volume'):
            try:
                del helper.tessellation_kwargs[_kw]
            except KeyError:
                pass
    elif tessellation_options is None:
        helper.tessellation_kwargs = kwargs
        helper.partition_kwargs = partition_options
    else:
        helper.tessellation_kwargs = tessellation_options
        helper.partition_kwargs = kwargs

    params = helper.standard_parameters(distance=distance, ref_distance=ref_distance, \
        rel_min_distance=rel_min_distance, rel_avg_distance=rel_avg_distance, \
        rel_max_distance=rel_max_distance, \
        min_location_count=min_location_count, avg_location_count=avg_location_count, \
        max_location_count=max_location_count, knn=knn, radius=radius, time_knn=time_knn)
    helper.parse_args(params, knn=knn, radius=radius, time_knn=time_knn, \
        rel_max_size=rel_max_size, rel_max_volume=rel_max_volume, \
        time_window_duration=time_window_duration, time_window_shift=time_window_shift, \
        time_window_options=time_window_options, enable_time_regularization=enable_time_regularization, \
        kwargs=kwargs)

    cells = helper.tessellate(comment=comment)
    cells.param.update(kwargs)

    if overwrite is None and force is not None:
        warn('`force` is deprecated; please use `overwrite` instead', PendingDeprecationWarning)
        overwrite = force
    if save_options is None:
        save_options = {}
    helper.save_analyses(output_file, force=overwrite, **save_options)

    if return_analyses:
        return helper.analyses
    else:
        return cells



fig_formats = ['png', 'pdf', 'ps', 'eps', 'svg', 'html']

def tessellate0(xyt_data, method='gwr', output_file=None, verbose=False, \
    scaling=False, time_scale=None, \
    knn=None, radius=None, distance=None, ref_distance=None, \
    rel_min_distance=None, rel_avg_distance=None, rel_max_distance=None, \
    min_location_count=None, avg_location_count=None, max_location_count=None, \
    rel_max_size=None, rel_max_volume=None, \
    time_window_duration=None, time_window_shift=None, time_window_options=None, \
    label=None, output_label=None, comment=None, input_label=None, inplace=False, \
    force=None, return_analyses=False, \
    load_options=None, tessellation_options=None, partition_options=None, save_options=None, \
    **kwargs):
    """
    Tessellation from points series and partitioning.

    This helper routine is a high-level interface to the various tessellation techniques
    implemented in TRamWAy.

    In addition to `knn`, `radius`, *filter* and *metric*, arguments with prefix *strict_* in
    their name apply to the partitioning step only, while the others apply to the tessellation step.

    *rel_max_size* and *rel_max_volume* are notable exceptions in that they currently apply to
    the partitioning step whereas they should conceptually apply to the tessellation step instead.
    This may change in a future version.

    Arguments:
        xyt_data (str or pandas.DataFrame):
            Path to a *.trxyt* or *.rwa* file or raw data in the shape of
            :class:`pandas.DataFrame`.


        method (str):
            Tessellation method or plugin name.
            See for example
            :class:`~tramway.tessellation.random.RandomMesh` ('*random*'),
            :class:`~tramway.tessellation.grid.RegularMesh` ('*grid*'),
            :class:`~tramway.tessellation.hexagon.HexagonalMesh` ('*hexagon*'),
            :class:`~tramway.tessellation.kdtree.KDTreeMesh` ('*kdtree*'),
            :class:`~tramway.tessellation.kmeans.KMeansMesh` ('*kmeans*') and
            :class:`~tramway.tessellation.gwr.GasMesh` ('*gas*' or '*gwr*').

        output_file (str):
            Path to a *.rwa* file. The resulting tessellation and data partition will be
            stored in this file. If `xyt_data` is a path to a file and `output_file` is not
            defined, then `output_file` will be adapted from `xyt_data` with extension
            *.rwa* and possibly overwrite the input file.

        verbose (bool or int): Verbose output.

        scaling (bool or str):
            Normalization of the data.
            Any of '*unitrange*', '*whiten*' or other methods defined in
            :mod:`tramway.core.scaler`.

        time_scale (bool or float):
            If this argument is defined and intepretable as ``True``, the time axis is
            scaled by this factor and used as a space variable for the tessellation (2D+T or
            3D+T, for example).
            This is equivalent to manually scaling the ``t`` column and passing
            ``scaling=True``.

        knn (int or pair of ints):
            After growing the tessellation, a minimum and maximum numbers of nearest
            neighbours of each cell center can be used instead of the entire cell
            population. Let us denote ``min_nn, max_nn = knn``. Any of ``min_nn`` and
            ``max_nn`` can be ``None``.
            If a single `int` is supplied instead of a pair, then `knn` becomes ``min_nn``.
            ``min_nn`` enables cell overlap and any point may be associated with several
            cells.
            See also :meth:`~tramway.tessellation.base.Delaunay.cell_index`.

        radius (float):
            After growing the tessellation as a set of centroids, a cell will consist of
            the locations within this distance from the centroid.
            See also :meth:`~tramway.tessellation.base.Delaunay.cell_index`.

        distance/ref_distance (float):
            Supposed to be the average translocation distance. Can be modified so that the
            cells are smaller or larger.

        rel_min_distance (float):
            Multiplies with `ref_distance` to define the minimum inter-cell distance.

        rel_avg_distance (float):
            Multiplies with `ref_distance` to define an upper on the average inter-cell
            distance.

        rel_max_distance (float):
            Multiplies with `ref_distance` to define the maximum inter-cell distance.

        min_location_count (int):
            Minimum number of points per cell. Depending on the method, can be strictly
            enforced or regarded as a recommendation.

        avg_location_count (int):
            Average number of points per cell. For non-plugin method, per default, it is
            set to four times `min_location_count`.

        max_location_count (int):
            Maximum number of points per cell. This is used by *kdtree* and *gwr*.

        rel_max_size (float):
            Maximum cell diameter as a number of `ref_distance`. Diameter (or size) is
            estimated as twice the distance between the center of cell and the nearest
            vertex. Cells of excess size are ignored so as the associated locations.

        rel_max_volume (float):
            Maximum cell volume (or surface area in 2D) as a number of `ref_distance`.
            Cells of excess volume are ignored so as the associated locations.

        strict_min_location_count (int):
            Minimum number of points per cell in the eventual partition. Cells with
            insufficient points are ignored so as the associated locations.

        strict_rel_max_size (float):
            Maximum cell diameter as a number of `ref_distance`. Diameter (or size) is
            estimated as the maximum distance between any pair of locations in the cell.
            Cells of excess size are ignored so as the associated locations.

        time_window_duration (float):
            Window duration in seconds (or frames with
            ``time_window_options=dict(frames=True)``).
            This time windowing combines with any other spatial tessellation method.
            To use the :mod:`~tramway.tessellation.window` plugin only, use ``method=window``
            and its *duration* and *shift* arguments instead.
            See also the :mod:`~tramway.tessellation.window` plugin.

        time_window_shift (float):
            Window shift in seconds (or frames with
            ``time_window_options=dict(frames=True)``).
            Default is no overlap, i.e. ``time_window_shift=time_window_duration``.
            See also the :mod:`~tramway.tessellation.window` plugin.

        time_window_options (dict):
            Extra arguments for time windowing.
            See also the :mod:`~tramway.tessellation.window` plugin.

        input_label (str):
            Label for the input tessellation for nesting tessellations.

        label/output_label (int or str):
            Label for the resulting analysis instance.

        inplace (bool):
            If True, `label`/`output_label`/`input_label` are exclusive, they all define
            a same analysis and the resulting analysis replaces the input analysis.

        comment (str):
            Description message for the resulting analysis.

        return_analyses (bool):
            Return a :class:`~tramway.core.analyses.base.Analyses` object instead of
            the default :class:`~tramway.tessellation.base.Partition` output.

        load_options (dict):
            Pass extra keyword arguments to :func:`~tramway.core.xyt.load_xyt` if called.

        tessellation_options (dict):
            Pass explicit keyword arguments to the *__init__* function of the
            tessellation class as well as to the
            :meth:`~tramway.tessellation.base.Tessellation.tessellate` method, and ignore
            the extra input arguments.

        partition_options (dict):
            Pass explicit keyword arguments to the
            :meth:`~tramway.tessellation.base.Tessellation.cell_index` method and ignore
            the extra input arguments.

        save_options (dict):
            Pass extra keyword arguments to :func:`~tramway.core.xyt.save_rwa` if called.

    Returns:
        tramway.tessellation.base.Partition: A partition of the data with its
            :attr:`~tramway.tessellation.base.Partition.tessellation` attribute set.


    Apart from the parameters defined above, extra input arguments are admitted and may be passed
    to the initializer of the selected tessellation method as well as to the
    :meth:`~tramway.tessellation.base.Tessellation.tessellate` and
    :meth:`~tramway.tessellation.base.Tessellation.cell_index` methods.

    See the individual documentation of these methods for more information.

    """
    if verbose:
        plugins.verbose = True

    compress = True
    if inplace:
        labels_exclusive = ValueError("multiple different values in exclusive arguments 'label', 'input_label' and 'output_label'")
        if label:
            if (input_label and input_label != label) or \
                (output_label and output_label != label):
                raise labels_exclusive
        elif input_label:
            if output_label and output_label == input_label:
                raise labels_exclusive
            label = input_label
        elif output_label:
            label = output_label
        output_label = input_label = label
    elif label is None:
        label = output_label
    elif output_label and output_label != label:
        raise ValueError("'label' and 'output_label' are both defined and are different")

    no_nesting_error = ValueError('nesting tessellations does not apply to translocation data')
    multiple_files = lambda a: isinstance(a, (tuple, list, frozenset, set))
    xyt_files = []
    if isinstance(xyt_data, six.string_types):
        # file path
        xyt_files = [xyt_data]
    elif multiple_files(xyt_data):
        # file path(s)
        xyt_files = list(xyt_data)
        if not xyt_data[1:]:
            xyt_data = xyt_data[0]
    if xyt_files:
        if load_options is None:
            load_options = {}
        if multiple_files(xyt_data):
            xyt_file = xyt_data
        else:
            try:
                analyses = load_rwa(xyt_data, lazy=True, verbose=False)
            except (KeyboardInterrupt, SystemExit):
                raise
            except:
                #traceback.print_exc()
                xyt_file = xyt_data
            else:
                xyt_file = False
                if input_label:
                    input_partition, = find_artefacts(analyses, Partition, input_label)
                xyt_data = analyses.data
        if xyt_file:
            xyt_data, xyt_files = load_xyt(xyt_files, return_paths=True, verbose=verbose,
                **load_options)
            analyses = Analyses(xyt_data)
            if input_label is not None:
                raise no_nesting_error
    else:
        if isinstance(xyt_data, abc.Analyses):
            analyses = xyt_data
            xyt_data = analyses.data
        else:
            analyses = Analyses(xyt_data)
        #warn('TODO: test direct data input', UseCaseWarning)
        if input_label is not None:
            raise no_nesting_error
    input_files = xyt_files

    try:
        setup, module = plugins[method]
        constructor = setup['make']
        if isinstance(constructor, str):
            constructor = getattr(module, setup['make'])
    except KeyError: # former code
        from tramway.tessellation.kdtree import KDTreeMesh
        from tramway.tessellation.kmeans import KMeansMesh
        from tramway.tessellation.gwr import GasMesh
        plugin = False
        methods = dict(grid=RegularMesh, kdtree=KDTreeMesh, kmeans=KMeansMesh, gwr=GasMesh)
        constructor = methods[method]
    else:
        plugin = True
    assert plugin

    transloc_length = min_distance = avg_distance = max_distance = None
    if ref_distance is None and distance is not None:
        # `distance` is only for compatibility with the tramway commandline
        ref_distance = distance
    if input_label and input_partition.param is not None:
        prm = input_partition.param.get('tessellation', input_partition.param)
        if not ref_distance:
            ref_distance = prm.get('ref_distance', None)
        min_distance = prm.get('min_distance', min_distance)
        avg_distance = prm.get('avg_distance', avg_distance)
        max_distance = prm.get('max_distance', max_distance)
    # former default values for `rel_min_distance` and `rel_avg_distance`
    if rel_min_distance is None and min_distance is None:
        rel_min_distance = .8
    if rel_avg_distance is None and avg_distance is None:
        rel_avg_distance = 2.

    if ref_distance is None:
        transloc_xy = np.asarray(translocations(xyt_data))
        if transloc_xy.shape[0] == 0:
            raise ValueError('no translocation found')
        transloc_length = np.nanmean(np.sqrt(np.sum(transloc_xy * transloc_xy, axis=1)))
        if verbose:
            print('average translocation distance: {}'.format(transloc_length))
        ref_distance = transloc_length
    if rel_min_distance is not None:
        min_distance = rel_min_distance * ref_distance
    if rel_avg_distance is not None:
        avg_distance = rel_avg_distance * ref_distance
    if rel_max_distance is not None:
        # applies only to KDTreeMesh and Kohonen
        max_distance = rel_max_distance * ref_distance
        #if method not in ['kdtree', 'gwr', 'kohonen']:
        #    warn('`rel_max_distance` is relevant only with `kdtree`', IgnoredInputWarning)

    if scaling:
        if scaling is True:
            scaling = 'whiten'
        scaler = dict(whiten=whiten, unit=unitrange)[scaling]()
    else:
        scaler = None

    if min_location_count is None: # former default value: 20
        if knn is None and radius is None:
            min_location_count = 20
    n_pts = float(xyt_data.shape[0])
    if min_location_count:
        min_probability = float(min_location_count) / n_pts
    else:
        min_probability = None
        if not plugin:
            warn('undefined `min_location_count`; not tested', UseCaseWarning)
    if not avg_location_count:
        if min_location_count is None:
            avg_location_count = 80 # former default value
        else:
            avg_location_count = 4 * min_location_count
    if avg_location_count:
        avg_probability = float(avg_location_count) / n_pts
    else:
        avg_probability = None
        if not plugin:
            warn('undefined `avg_location_count`; not tested', UseCaseWarning)
    if max_location_count:
        # applies only to KDTreeMesh
        max_probability = float(max_location_count) / n_pts
        if not plugin and method != 'kdtree':
            warn('`max_location_count` is relevant only with `kdtree`', IgnoredInputWarning)
    else:
        max_probability = None

    # actually useful only if no tessellation nesting
    colnames = ['x', 'y']
    if 'z' in xyt_data:
        colnames.append('z')
    if time_scale:
        colnames.append('t')
        if scaler is None:
            raise ValueError('no scaling defined')
        scaler.factor = [('t', time_scale)]

    # distinguish between tessellation and partition arguments
    if tessellation_options is None and partition_options is None:
        tessellation_kwargs = dict(kwargs)
        partition_kwargs = {}
        for _kw in kwargs:
            if _kw.startswith('strict_'):
                partition_kwargs[_kw[7:]] = tessellation_kwargs.pop(_kw)
        for _kw in ('filter', 'filter_descriptors_only', 'metric'):
            try:
                _arg = tessellation_kwargs.pop(_kw)
            except KeyError:
                pass
            else:
                partition_kwargs[_kw] = _arg
        for _kw in ('rel_max_size', 'rel_max_volume'):
            try:
                del tessellation_kwargs[_kw]
            except KeyError:
                pass
    elif tessellation_options is None:
        tessellation_kwargs = kwargs
        partition_kwargs = partition_options
    else:
        tessellation_kwargs = tessellation_options
        partition_kwargs = kwargs

    # initialize a Tessellation object
    params = dict( \
        min_distance=min_distance, \
        avg_distance=avg_distance, \
        max_distance=max_distance, \
        min_probability=min_probability, \
        avg_probability=avg_probability, \
        max_probability=max_probability, \
        )
    if plugin:
        for ignored in ['max_level']:
            try:
                if tessellation_kwargs[ignored] is None:
                    del tessellation_kwargs[ignored]
            except KeyError:
                pass
        params.update(dict( \
            min_location_count=min_location_count, \
            avg_location_count=avg_location_count, \
            max_location_count=max_location_count, \
            ))
        params.update(tessellation_kwargs)
        for key in setup.get('make_arguments', {}):
            try:
                param = params[key]
            except KeyError:
                pass
            else:
                tessellation_kwargs[key] = param
    else:
        params.update(tessellation_kwargs)
        tessellation_kwargs = params
    time_window_kwargs = {}

    if input_label:
        if time_window_duration:
            raise NotImplementedError('spatial tessellation combined with time windowing is not supported yet by tessellation nesting')
        tess = NestedTessellations(scaler, input_partition, factory=constructor,
            **tessellation_kwargs)
        xyt_data = data = input_partition.points
    else:
        if time_window_duration:
            time_window_kwargs['duration'] = time_window_duration
            if time_window_shift:
                time_window_kwargs['shift'] = time_window_shift
            if time_window_options:
                time_window_kwargs.update(time_window_options)
            import tramway.tessellation.window as window
            tess = window.SlidingWindow(**time_window_kwargs)
            tess.spatial_mesh = constructor(scaler, **tessellation_kwargs)
        else:
            tess = constructor(scaler, **tessellation_kwargs)
        data = xyt_data[colnames]

    # grow the tessellation
    tessellate_kwargs = tessellation_kwargs
    tess.tessellate(data, verbose=verbose, **tessellate_kwargs)

    # partition the dataset into the cells of the tessellation
    _filter_f = partition_kwargs.get('filter', None)
    try:
        max_size = kwargs['rel_max_size']
    except KeyError:
        _filter_fg = _filter_f
    else:
        max_size *= ref_distance
        def _filter_g(voronoi, cell, points):
            return voronoi.scaler.unscale_distance(np.sqrt(np.min(np.sum(( \
                    voronoi._center[cell] - \
                    voronoi._vertices[voronoi.cell_vertices[cell]] \
                )*2, axis=1)))) <= max_size
        _filter_fg = _filter_g if _filter_f is None \
            else lambda *a: _filter_g(*a) and _filter_f(*a)
    try:
        max_volume = kwargs['rel_max_volume']
    except KeyError:
        _filter_fgh = _filter_fg
    else:
        max_volume *= ref_distance
        def _filter_h(voronoi, cell, points):
            return voronoi.cell_volume[cell] <= max_volume
        _filter_fgh = _filter_h if _filter_fg is None \
            else lambda *a: _filter_h(*a) and _filter_fg(*a)
    try:
        max_actual_size = partition_kwargs.pop('rel_max_size')
    except KeyError:
        _filter_fghi = _filter_fgh
    else:
        max_actual_size *= ref_distance
        _descr = not partition_kwargs.get('filter_descriptors_only', True)
        def _filter_i(voronoi, cell, x):
            if _descr:
                x = voronoi.descriptors(x, asarray=True)
            x2 = np.sum(x * x, axis=1, keepdims=True)
            d2 = x2 + x2.T - 2. * np.dot(x, x.T)
            d = np.sqrt(np.max(d2.ravel()))
            return d <= max_actual_size
        _filter_fghi = _filter_i if _filter_fgh is None \
            else lambda *a: _filter_i(*a) and _filter_fgh(*a)
    if _filter_fghi is not None:
        partition_kwargs['filter'] = _filter_fghi
        if 'filter_descriptors_only' not in partition_kwargs:
            partition_kwargs['filter_descriptors_only'] = True

    if not (knn is None and radius is None):
        if knn is not None:
            partition_kwargs['knn'] = knn
        if radius is not None:
            if 'radius' in partition_kwargs:
                warn('overwriting `radius`', RuntimeWarning)
            partition_kwargs['radius'] = radius
        if 'min_location_count' not in partition_kwargs:
            partition_kwargs['min_location_count'] = min_location_count
        if 'metric' not in partition_kwargs:
            partition_kwargs['metric'] = 'euclidean'

    try:
        cell_index = tess.cell_index(xyt_data, **partition_kwargs)
    except MemoryError:
        if verbose:
            print(traceback.format_exc())
        warn('memory error: cannot assign points to cells', RuntimeWarning)
        cell_index = None

    stats = Partition(xyt_data, tess, cell_index)

    # store some parameters together with the partition
    stats.param['method'] = method
    #if transloc_length:
    #    stats.param['transloc_length'] = transloc_length
    #else:
    #    stats.param['ref_distance'] = ref_distance
    #for _arg in ('min_distance', 'avg_distance', 'max_distance', 'knn', 'radius', \
    #        'time_window_duration', 'time_window_shift', 'time_window_options'):
    #    _val = eval(_arg)
    #    if _val:
    #        stats.param[_arg] = _val
    if time_window_kwargs:
        stats.param['time_window'] = time_window_kwargs
    if tessellation_kwargs:
        stats.param['tessellation'] = tessellation_kwargs
    if partition_kwargs:
        stats.param['partition'] = partition_kwargs
    if reassignment_kwargs:
        stats.param['reassignment'] = reassignment_kwargs
    stats.param.update(kwargs)

    # insert the resulting analysis in the analysis tree
    if input_label:
        input_analysis = analyses
        if isinstance(input_label, (tuple, list)):
            prefix_labels = list(input_label)
            terminal_label = prefix_labels.pop()
            for _label in prefix_labels:
                input_analysis = input_analysis.instances[_label]
        else:
            terminal_label = input_label
        # `input_analysis` is already in `analyses`
        if inplace and comment:
            input_analysis.comments[terminal_label] = comment
        input_analysis = input_analysis.instances[terminal_label]
        if inplace:
            input_analysis.artefact = stats
        else:
            input_analysis.add(Analyses(stats), label=label, comment=comment)
    else:
        analyses.add(Analyses(stats), label=label, comment=comment)

    # save the analysis tree (`analyses`)
    if output_file or xyt_files:
        if output_file is None:
            output_file = os.path.splitext(xyt_files[0])[0] + '.rwa'
        if save_options is None:
            save_options = {}
        if 'force' not in save_options:
            save_options['force'] = force or (force is not False and len(input_files)==1 and input_files[0]==output_file)

        save_rwa(output_file, analyses, verbose, **save_options)

    if return_analyses:
        return analyses
    else:
        return stats




def cell_plot(cells, xy_layer=None, output_file=None, fig_format=None, \
    show=None, verbose=False, figsize=None, dpi=None, \
    location_count_hist=False, cell_dist_hist=False, location_dist_hist=False, \
    aspect=None, delaunay=None, locations={}, voronoi=None, colors=None, title=None, \
    cell_indices=None, segment=None, label=None, input_label=None, num = None, \
    **kwargs):
    """
    Partition plots.

    Plots a spatial representation of the tessellation and partition if data are 2D.

    Arguments:
        cells (str or Partition or tramway.core.analyses.base.Analyses):
            Path to a *.rwa* file or :class:`~tramway.tessellation.Partition`
            instance or analysis tree; files and analysis trees may require
            `label`/`input_label` to be defined.

        xy_layer ({None, 'delaunay', 'voronoi'}):
            Overlay Delaunay or Voronoi graph over the data points. For 2D data only.
            **Deprecated**! Please use `delaunay` and `voronoi` arguments instead.

        output_file (str):
            Path to a file in which the figure will be saved. If `cells` is a path and
            `fig_format` is defined, `output_file` is automatically set.

        fig_format (str):
            Any image format supported by :func:`matplotlib.pyplot.savefig`.

        show (bool or str):
            Makes `cell_plot` show the figure(s) which is the default behavior if and only
            if the figures are not saved.
            If ``show='draw'``, :func:`~matplotlib.pyplot.draw` is called instead of
            :func:`~matplotlib.pyplot.show`.
            To maintain the current default behaviour in the future, set `show` to
            ``True`` from now on.

        verbose (bool): Verbose output.

        figsize (pair of floats):
            Passed to :func:`matplotlib.pyplot.figure`. Applies only to the spatial
            representation figure.

        dpi (int):
            Passed to :func:`matplotlib.pyplot.savefig`. Applies only to the spatial
            representation figure.

        location_count_hist (bool):
            Plot a histogram of point counts (per cell). If the figure is saved, the
            corresponding file will have sub-extension *.hpc*;
            **deprecated**.

        cell_dist_hist (bool):
            Plot a histogram of distances between neighbour centroids. If the figure is
            saved, the corresponding file will have sub-extension *.hcd*;
            **deprecated**.

        location_dist_hist (bool):
            Plot a histogram of distances between points from neighbour cells. If the figure
            is saved, the corresponding file will have sub-extension *.hpd*;
            **deprecated**.

        aspect (str):
            Aspect ratio. Can be '*equal*'.

        locations (dict):
            Keyword arguments to :func:`~tramway.plot.mesh.plot_points`.

        delaunay (bool or dict):
            Overlay Delaunay graph. If :class:`dict`, keyword arguments to
            :func:`~tramway.plot.mesh.plot_delaunay`.

        voronoi (bool or dict):
            Overlay Voronoi graph (default: ``True``). If :class:`dict`, keyword arguments
            to :func:`~tramway.plot.mesh.plot_voronoi`.

        cell_indices (bool or dict):
            ``True`` or keyworded arguments; plot cell indices instead of centroids.

        label/input_label (int or str or list):
            If `cells` is a filepath or an analysis tree, label of the analysis instance.

        segment (int):
            Segment index; if multiple time segments were defined, show only one segment.

        num (int):
            Figure number. Default: None (new figure for each call).

    Notes:
        See also :mod:`tramway.plot.mesh`.

    """
    input_file = ''
    if not isinstance(cells, Partition):
        if label is None:
            labels = input_label
        else:
            labels, label = label, None
        if not valid_label(labels):
            labels = ()
        elif not isinstance(labels, (tuple, list)):
            labels = (labels, )
        if isinstance(cells, abc.Analyses):
            analyses, cells = cells, None
        else:
            input_file, cells = cells, None
            if isinstance(input_file, (tuple, list)):
                if input_file[1:]:
                    warn('can only process a single file', RuntimeWarning)
                input_file = input_file[0]
            if not isinstance(input_file, str):
                raise TypeError('unsupported input data type: {}'.format(type(input_file)))
            try:
                analyses = load_rwa(input_file, lazy=True)
                #if labels:
                #       analyses = extract_analysis(analyses, labels)
            except KeyError as e:
                if e.args and 'analyses' not in e.args[0]:
                    raise
                # legacy code
                imt_path = input_file
                if imt_path is None:
                    raise ValueError('undefined input file')
                # copy-paste
                if os.path.isdir(imt_path):
                    imt_path = os.listdir(imt_path)
                    files, exts = zip(*os.path.splitext(imt_path))
                    for ext in imt_extensions:
                        if ext in exts:
                            imt_path = imt_path[exts.index(ext)]
                            break
                    if isinstance(imt_path, list):
                        imt_path = imt_path[0]
                    auto_select = True
                elif os.path.isfile(imt_path):
                    auto_select = False
                else:
                    candidates = [ imt_path + ext for ext in imt_extensions ]
                    candidates = [ f for f in candidates if os.path.isfile(f) ]
                    if candidates:
                        imt_path = candidates[0]
                    else:
                        raise IOError('no tessellation file found in {}'.format(imt_path))
                    auto_select = True
                if auto_select and verbose:
                    print('selecting {} as a tessellation file'.format(imt_path))

                # load the data
                input_file = imt_path
                try:
                    from rwa import HDF5Store
                    hdf = HDF5Store(input_file, 'r')
                    hdf.lazy = False
                    cells = hdf.peek('cells')
                    if cells.tessellation is None:
                        cells._tessellation = hdf.peek('_tesselation', hdf.store['cells'])
                except:
                    print(traceback.format_exc())
                    warn('HDF5 libraries may not be installed', ImportWarning)
                finally:
                    hdf.close()
        if cells is None:
            if isinstance(analyses, dict):
                if not analyses:
                    raise ValueError('not any file matches')
                elif len(analyses) == 1:
                    analyses = list(analyses.values())[0]
                else:
                    raise ValueError('multiple files match')
            if not labels:
                labels = list(analyses.labels)
            if labels[1:]:
                raise ValueError('multiple instances; label is required')
            label = labels[-1]
            cells = analyses[label].data

    # identify time segments if any
    try:
        import tramway.tessellation.time as time
        with_segments = isinstance(cells.tessellation, time.TimeLattice) \
                and cells.tessellation.spatial_mesh is not None
    except ImportError:
        with_segments = False
    if with_segments:
        if segment is None:
            xyt = cells.points
            mesh = cells.tessellation.spatial_mesh
            prms = cells.param.get('partition', {})
            prms.pop('exclude_cells_by_location_count', None)
            time_col = prms.pop('time_col', 't')
            try:
                prms.pop('time_knn')
            except KeyError:
                pass
            else:
                warn('ignoring `time_knn`', RuntimeWarning)
            bb = cells.bounding_box
            cells = Partition(xyt, mesh, mesh.cell_index(xyt, **prms))
            cells.bounding_box = bb
        else:
            if isinstance(segment, (tuple, list)):
                if segment[1:]:
                    warn('cannot plot multiple segments in a single `cell_plot` call', RuntimeWarning)
                segment = segment.pop()
                print('plotting segment {}'.format(segment))
            cells = cells.tessellation.split_segments(cells)[segment]
    elif segment is not None:
        warn('cannot find time segments', RuntimeWarning)
        segment = None

    # guess back some input parameters (for backward compatible titles)
    complementary_plots = location_count_hist or cell_dist_hist or location_dist_hist
    if (title and not isinstance(title, str)) or complementary_plots:
        method_name = {}
        try:
            method_name[RegularMesh] = ('grid', 'grid', 'regular grid')
        except NameError:
            pass
        try:
            import tramway.tessellation.kdtree
        except ImportError:
            pass
        else:
            method_name[tramway.tessellation.kdtree.KDTreeMesh] = \
                ('kdtree', 'k-d tree', 'k-d tree based tessellation')
        try:
            import tramway.tessellation.kmeans
        except ImportError:
            pass
        else:
            method_name[tramway.tessellation.kmeans.KMeansMesh] = \
                ('kmeans', 'k-means', 'k-means based tessellation')
        try:
            import tramway.tessellation.gwr
        except ImportError:
            pass
        else:
            method_name[tramway.tessellation.gwr.GasMesh] = ('gwr', 'GWR', 'GWR based tessellation')
        try:
            method_name, pp_method_name, method_title = method_name[type(cells.tessellation)]
        except (KeyError, AttributeError):
            method_name = pp_method_name = method_title = ''

    if complementary_plots:
        warn('complementary plots will be removed in a future release', DeprecationWarning)
        try:
            min_distance = kwargs['min_distance']
        except KeyError:
            min_distance = cells.param['tessellation'].get('min_distance', 0)
        try:
            avg_distance = kwargs['avg_distance']
        except KeyError:
            avg_distance = cells.param['tessellation'].get('avg_distance', None)
    try:
        min_location_count = kwargs['min_location_count']
    except KeyError:
        try:
            min_location_count = cells.param['partition']['min_location_count']
        except (KeyError, AttributeError):
            min_location_count = 0

    print_figs = output_file or (input_file and fig_format)
    if (print_figs and figsize is None) or figsize is True:
        figsize = (12., 9.)

    # import graphics libraries with adequate backend
    if print_figs:
        import matplotlib
        try:
            matplotlib.use('Agg') # head-less rendering (no X server required)
        except:
            pass
    import matplotlib.pyplot as mplt
    import tramway.plot.mesh as tplt

    # plot the data points together with the tessellation
    figs = []
    try:
        dim = cells.tessellation.cell_centers.shape[1]
    except AttributeError as e:
        try:
            dim = cells.dim
        except AttributeError:
            raise e
    fig = None
    if dim == 2 and not complementary_plots:
        if 'figure' in kwargs:
            fig = kwargs['figure']
        elif figs or figsize is not None or num is not None:
            fig = mplt.figure(figsize=figsize, dpi=dpi, num = num)
        else:
            fig = mplt.gcf()
        figs.append(fig)
        if locations is not None:
            if 'axes' in kwargs:
                locations['axes'] = kwargs['axes']
            if 'knn' in cells.param: # if knn <= min_count, min_count is actually ignored
                tplt.plot_points(cells, **locations)
            else:
                tplt.plot_points(cells, min_count=min_location_count, **locations)
        if aspect is not None:
            ax = kwargs.get('axes', None)
            if ax is None:
                ax = fig.gca()
            ax.set_aspect(aspect)
        if voronoi is None:
            voronoi = issubclass(type(cells.tessellation), Voronoi)
        if xy_layer != 'delaunay' and voronoi:
            if not isinstance(voronoi, dict):
                voronoi = {}
            if cell_indices and 'centroid_style' not in voronoi:
                voronoi['centroid_style'] = None
            if 'axes' in kwargs:
                voronoi['axes'] = kwargs['axes']
            tplt.plot_voronoi(cells, **voronoi)
            voronoi = True
        if xy_layer == 'delaunay' or delaunay: # Delaunay above Voronoi
            if not isinstance(delaunay, dict):
                delaunay = {}
            if 'axes' in kwargs:
                delaunay['axes'] = kwargs['axes']
            tplt.plot_delaunay(cells, **delaunay)
            delaunay = True
        if cell_indices:
            if not isinstance(cell_indices, dict):
                cell_indices = {}
            if 'axes' in kwargs:
                cell_indices['axes'] = kwargs['axes']
            tplt.plot_indices(cells, **cell_indices)
        if title:
            if isinstance(title, str):
                _title = title
            elif delaunay == voronoi:
                _title = pp_method_name
            elif delaunay:
                _title = pp_method_name + ' based Delaunay'
            elif voronoi:
                _title = pp_method_name + ' based Voronoi'
            try:
                axes = kwargs['axes']
            except KeyError:
                mplt.title(_title)
            else:
                axes.set_title(_title)


    if print_figs:
        if output_file:
            filename, figext = os.path.splitext(output_file)
            if fig_format:
                figext = fig_format
            elif figext and figext[1:] in fig_formats:
                figext = figext[1:]
            else:
                figext = fig_formats[0]
        else:
            figext = fig_format
            filename, _ = os.path.splitext(input_file)
        subname, subext = os.path.splitext(filename)
        if subext and subext[1:] in ['imt']: # very old file format
            filename = subname
        if fig is not None:
            vor_file = '{}.{}'.format(filename, figext)
            if verbose:
                print('writing file: {}'.format(vor_file))
            fig.savefig(vor_file, dpi=dpi)


    # the complementary histograms below haven't been tested for a while [TODO]

    if location_count_hist:
        # plot a histogram of the number of points per cell
        fig = mplt.figure()
        figs.append(fig)
        mplt.hist(cells.location_count, bins=np.arange(0, min_location_count*20, min_location_count))
        mplt.plot((min_location_count, min_location_count), mplt.ylim(), 'r-')
        mplt.title(method_title)
        mplt.xlabel('point count (per cell)')
        if print_figs:
            hpc_file = '{}.{}.{}'.format(filename, 'hpc', figext)
            if verbose:
                print('writing file: {}'.format(hpc_file))
            fig.savefig(hpc_file)

    if cell_dist_hist:
        # plot a histogram of the distance between adjacent cell centers
        A = sparse.triu(cells.tessellation.cell_adjacency, format='coo')
        i, j, k = A.row, A.col, A.data
        label = cells.tessellation.adjacency_label
        if label is not None:
            i = i[0 < label[k]]
            j = j[0 < label[k]]
        pts = cells.tessellation.cell_centers
        dist = pts[i,:] - pts[j,:]
        dist = np.sqrt(np.sum(dist * dist, axis=1))
        fig = mplt.figure()
        figs.append(fig)
        mplt.hist(np.log(dist), bins=50)
        if avg_distance:
            dmin = np.log(min_distance)
            dmax = np.log(avg_distance)
            ylim = mplt.ylim()
            mplt.plot((dmin, dmin), ylim, 'r-')
            mplt.plot((dmax, dmax), ylim, 'r-')
        mplt.title(method_title)
        mplt.xlabel('inter-centroid distance (log)')
        if print_figs:
            hcd_file = '{}.{}.{}'.format(filename, 'hcd', figext)
            if verbose:
                print('writing file: {}'.format(hcd_file))
            fig.savefig(hcd_file)

    if location_dist_hist:
        adj = point_adjacency_matrix(cells, symetric=False)
        dist = adj.data
        fig = mplt.figure()
        figs.append(fig)
        mplt.hist(np.log(dist), bins=100)
        if avg_distance:
            dmin = np.log(min_distance)
            dmax = np.log(avg_distance)
            ylim = mplt.ylim()
            mplt.plot((dmin, dmin), ylim, 'r-')
            mplt.plot((dmax, dmax), ylim, 'r-')
        mplt.title(method_title)
        mplt.xlabel('inter-point distance (log)')
        if print_figs:
            hpd_file = '{}.{}.{}'.format(filename, 'hpd', figext)
            if verbose:
                print('writing file: {}'.format(hpd_file))
            fig.savefig(hpd_file)

    if show or not print_figs: # 'or' will become 'and'
        if show == 'draw':
            mplt.draw()
        elif show is not False:
            mplt.show()
    else:
        for fig in figs:
            mplt.close(fig)


def delete_low_count_cells(partition, count_threshold, priority_by=None, label=True, partition_kwargs={}):
    tessellation = partition.tessellation
    deleted_cells, = np.nonzero(partition.location_count<count_threshold)
    #print('ncells', partition.number_of_cells, 'npts_min', np.min(partition.location_count), 'npts_max', np.max(partition.location_count), 'ncells_deleted', deleted_cells.size)
    if deleted_cells.size == 0:
        return partition, deleted_cells, label
    if priority_by:
        if priority_by == 'count':
            priority = -partition.location_count[deleted_cells]
        elif priority_by == 'volume':
            priority = tessellation.volume[deleted_cells]
        ordering = np.argsort(priority)
        deleted_cells = deleted_cells[ordering]
        index_mapping, label = tessellation.delete_cells(deleted_cells, exclude_neighbours=True, adjacency_label=label)
    else:
        index_mapping, label = tessellation.delete_cells(deleted_cells)

    points = partition.points
    cell_indices = tessellation.cell_index(points, **partition_kwargs)
    new_partition = Partition(points, tessellation, cell_indices)
    return new_partition, deleted_cells, label


def update_cell_centers(cells, max_iter, partition_kwargs={}):
    points = cells.points[['x','y']].values # TODO: use `descriptors` instead
    tess = cells.tessellation
    cell_indices = cells.cell_index

    if max_iter is True:
        max_iter = np.inf

    k = 0
    while k < max_iter:
        prev_cell_indices = cell_indices
        cell_centers = np.array(tess.cell_centers) # copy
        any_new = False
        for i in range(tess.number_of_cells):
            prev_cell_i = prev_cell_indices==i
            cell_i = cell_indices==i
            if np.any(cell_i) and not np.all(cell_i==prev_cell_i):
                any_new = True
                center = np.mean(points[cell_i], axis=0)
                assert not np.any(np.isnan(center))
                cell_centers[i] = center
        if any_new:
            tess.cell_centers = cell_centers
            cell_indices = tess.cell_index(cells.points, **partition_kwargs)
            k += 1
        else:
            break
    return Partition(cells.points, tess, cell_indices)


tessellate = tessellate1




from . import *
from ..attribute import initializer_method

from_plugin   = initializer_method( MapperInitializer.from_plugin   )
from_callable = initializer_method( MapperInitializer.from_callable )




# -*- coding: utf-8 -*-

# Copyright © 2017-2018, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from tramway.core import rc
from rwa import HDF5Store, lazytype, lazyvalue, islazy
from ..lazy import Lazy
from ..analyses import Analyses, coerce_labels, format_analyses, append_leaf
import tramway.core.analyses.abc as abc
import os.path
import traceback
import errno

try:
    input = raw_input # Py2
except NameError:
    pass


__all__ = ['RWAStore', 'load_rwa', 'save_rwa']


class RWAStore(HDF5Store):

    __slots__ = ('unload', '__special__', 'force_load_special')

    def __init__(self, resource, mode='auto', unload=False, verbose=False, **kwargs):
        HDF5Store.__init__(self, resource, mode, verbose, **kwargs)
        self.unload = unload
        self.__special__ = {}
        self.force_load_special = False

    def poke(self, objname, obj, container=None, visited=None, _stack=None, unload=None):
        if unload is not None:
            previous = self.unload
            self.unload = unload
        if visited is None:
            visited = {}
        if container is None:
            container = self.store
        try:
            obj = obj.lazy_unwrap()
        except AttributeError:
            pass
        obj = lazyvalue(obj, deep=True) # in the case it was stored not unloaded
        if self.unload and isinstance(obj, Lazy):
            # set lazy attributes to None (unset them so that memory is freed)
            #obj = copy(obj) # this causes a weird bug
            for name in obj.__lazy__:
                # `_lazy` should contain `name`;
                # if it is not, the object definition has changed
                if obj._lazy.get(name, True):
                    setattr(obj, obj.__fromlazy__(name), None)
        obj = self.special_unload(obj)
        HDF5Store.poke(self, objname, obj, container, visited=visited, _stack=_stack)
        if unload is not None:
            self.unload = previous

    def peek(self, objname, record=None, lazy=None, **kwargs):
        obj = HDF5Store.peek(self, objname, record=record, lazy=lazy, **kwargs)
        obj = self.special_load(obj, objname, lazy=lazy)
        return obj

    def special_load(self, obj, objname, lazy=None):
        if objname != '_data':
            return obj
        if 'data0' in self.__special__:
            import tramway.tessellation.base as tessellation
            if issubclass(lazytype(obj), tessellation.Partition):
                obj = lazyvalue(obj, deep=True)
                try:
                    if obj._points is None:
                        raise AttributeError
                except AttributeError:
                    obj._points = lazyvalue(self.__special__['data0'], deep=True)
        else:
            import pandas
            if issubclass(lazytype(obj), pandas.DataFrame):
                if self.force_load_special or not (self.lazy if lazy is None else lazy):
                    obj = lazyvalue(obj, deep=True)
                self.__special__['data0'] = obj
        return obj

    def special_unload(self, obj):
        # forced poke (calling lazyvalue) may not be necessary
        if 'data0' in self.__special__:
            import tramway.tessellation.base as tessellation
            if issubclass(lazytype(obj), tessellation.Partition):
                obj = lazyvalue(obj, deep=True)
                if obj.points is self.__special__['data0']:
                    import copy
                    obj = copy.copy(obj)
                    obj._points = None
        else:
            import pandas
            if issubclass(lazytype(obj), pandas.DataFrame):
                obj = lazyvalue(obj, deep=True)
                self.__special__['data0'] = obj
        return obj




def load_rwa(path, verbose=None, lazy=False, force_load_spt_data=None):
    """
    Load a .rwa file.

    Note about laziness: the analysis tree uses an active handle to the opened file.
    As a consequence, the file should be read only once.
    It is safe to load, modify, save and then load again, but the first loaded data
    should be terminated before loading the data again.

    Arguments:

        path (str): path to .rwa file

        verbose (bool or int): verbosity level

        lazy (bool): reads the file lazily

        force_load_spt_data (bool): *new in 0.5*
            compatibility flag for pre-0.5 code;
            `None` currently defaults to `True`,
            but will default to `False` in the future

    Returns:

        tramway.core.analyses.base.Analyses:
            analysis tree;
            if `lazy` is ``True``, return type is
            :class:`tramway.core.analyses.lazy.Analyses` instead
    """
    try:
        hdf = RWAStore(path, 'r', verbose=max(0, int(verbose) - 2) if verbose else False)
        #hdf._default_lazy = PermissivePeek
        hdf.lazy = lazy
        hdf.force_load_special = True if force_load_spt_data is None else force_load_spt_data
        try:
            analyses = lazyvalue(hdf.peek('analyses'))
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            if verbose is not False:
                print('cannot load file: {}'.format(path))
            raise
        finally:
            if not lazy:
                hdf.close()
    except EnvironmentError as e:
        if hasattr(e, 'errno') and e.errno == errno.ENOENT:
            raise
        elif e.args[1:]:
            if verbose:
                print(traceback.format_exc())
            raise OSError('HDF5 libraries may not be installed')
        else:
            raise
    return analyses
    return coerce_labels(analyses)



def save_rwa(path, analyses, verbose=False, force=None, compress=True, append=False, overwrite=None):
    """
    Save an analysis tree into a .rwa file.

    Arguments:

        path (str): path to .rwa file

        analyses (tramway.core.analyses.base.Analyses): analysis tree

        verbose (bool or int): verbose mode

        force/overwrite (bool): do not ask whether to overwrite an existing file or not

        compress (bool): delete the lazy attributes that can be computed again automatically

        append (bool): do not overwrite; reload the file instead and append the analyses as
            a subtree

    """
    if not isinstance(analyses, abc.Analyses):
        raise TypeError('`analyses` is not an `Analyses` instance')
    if force is None:
        force = False if overwrite is None else overwrite
    if os.path.isfile(path):
        if append:
            extra_analyses = analyses
            analyses = load_rwa(path)
            append_leaf(analyses, extra_analyses, overwrite=force)
        elif not force and rc.__user_interaction__:
            answer = input("overwrite file '{}': [N/y] ".format(path))
            if not (answer and answer[0].lower() == 'y'):
                return
    elif append:
        if verbose:
            print('file not found; flushing all the analyses')
    try:
        store = RWAStore(path, 'w', verbose=max(0, int(verbose) - 2))
        try:
            store.unload = compress
            if verbose:
                print('writing file: {}'.format(path))
            store.poke('analyses', analyses)
        finally:
            store.close()
    except EnvironmentError as e:
        if hasattr(e, 'errno') and e.errno == errno.ENOENT:
            raise
        print(traceback.format_exc())
        raise ImportError('HDF5 libraries may not be installed')
    if 1 < int(verbose):
        print('written analysis tree:')
        print(format_analyses(analyses, global_prefix='\t', node=lazytype))





# -*- coding: utf-8 -*-

# Copyright © 2020-2021, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from tramway.core.exceptions import MisplacedAttributeWarning, SideEffectWarning
import warnings
import logging

def report_misplaced_attribute(attr_name, proper_parent_name):
    warnings.warn(f'`{attr_name}` is an attribute of the initialized `{proper_parent_name}` attribute; this warning message can safely be silenced', MisplacedAttributeWarning)
def proper_parent_name(attr_name):
    parent_name = None
    get_conditions, set_conditions = {}, {}
    set_conditions['initialized'] = True
    if attr_name in ('dt', 'time_step', 'frame_interval',
            'localization_error', 'localization_precision', 'columns',
            'temperature'):
        parent_name = 'spt_data'
    elif attr_name in ('scaler', 'resolution'):
        parent_name = 'tesseller'
    elif attr_name in ('script',):
        parent_name = 'env'
    return parent_name, get_conditions, set_conditions

warnings.filterwarnings('error', category=SideEffectWarning)

_emulate_undefined_from_callable = False


from .attribute import *
from .artefact  import *
from .spt_data  import *
from .roi       import *
from .time      import *
from .tesseller import *
from .sampler   import *
from .mapper    import *
from .env       import *
from .pipeline  import *
from .browser   import *
from .images    import *
from .localizer import *
from .tracker   import *


import importlib
class AttributeSubPackage(object):
    __slots__ = ('__attrname__', '_module')
    def __init__(self, attrname):
        self.__attrname__ = attrname
        self.module = None
    @property
    def module(self):
        if self._module is None:
            self._module = importlib.import_module(
                    '.{}.allsymbols'.format(self.__attrname__),
                    package='tramway.analyzer')
        return self._module
    @module.setter
    def module(self, mod):
        self._module = mod
    def __getattr__(self, attrname):
        return getattr(self.module, attrname)

spt_data  = AttributeSubPackage('spt_data' )
roi       = AttributeSubPackage('roi'      )
time      = AttributeSubPackage('time'     )
tesseller = AttributeSubPackage('tesseller')
sampler   = AttributeSubPackage('sampler'  )
mapper    = AttributeSubPackage('mapper'   )
images    = AttributeSubPackage('images'   )
tracker   = AttributeSubPackage('tracker'  )


class BasicLogger(object):
    """
    Emulates the most basic functionalities of :class:`logging.Logger`
    without the need for additional configuration.
    """
    __slots__ = ('_level',)
    def __init__(self, level=0):
        self.level = level
    def _print(self, msg, lvl):
        if self.level <= lvl:
            print(msg)
    def debug(self, msg):
        self._print(msg, 10)
    def info(self, msg):
        self._print(msg, 20)
    def warning(self, msg):
        self._print(msg, 30)
    def error(self, msg):
        self._print(msg, 40)
    def critical(self, msg):
        self._print(msg, 50)
    @property
    def level(self):
        return self._level
    @level.setter
    def level(self, lvl):
        if isinstance(lvl, str):
            lvl = lvl.lower()
            lvl = dict(
                    notset = 0,
                    debug = 10,
                    info = 20,
                    warning = 30,
                    error = 40,
                    critical = 50,
                ).get(lvl, 0)
        elif not isinstance(lvl, int):
            try:
                lvl = int(lvl)
            except:
                lvl = 0
        self._level = lvl
    def setLevel(self, lvl):
        self.level = lvl


module_logger = logging.getLogger(__name__)
if not module_logger.hasHandlers():
    module_logger.setLevel(logging.INFO)
    module_logger.addHandler(logging.StreamHandler())


class RWAnalyzer(WithLogger):
    """
    A :class:`RWAnalyzer` object gathers the parameters of all the processing steps
    of a standard processing chain, from SPT data loading/generation to
    inferring model parameters at microdomains.

    The supported steps are defined in a declarative way with special attributes;
    these steps and corresponding attributes are as follows:

    * :attr:`~tramway.analyzer.RWAnalyzer.images`: microscopy images
    * :attr:`~tramway.analyzer.RWAnalyzer.localizer`: single molecule localization
    * :attr:`~tramway.analyzer.RWAnalyzer.tracker`: single particle tracking
    * :attr:`~tramway.analyzer.RWAnalyzer.spt_data`: SPT data loading or generation
    * :attr:`~tramway.analyzer.RWAnalyzer.roi`: regions of interest
    * :attr:`~tramway.analyzer.RWAnalyzer.tesseller`: spatial segmentation
    * :attr:`~tramway.analyzer.RWAnalyzer.time`: temporal segmentation of the tracking data
    * :attr:`~tramway.analyzer.RWAnalyzer.sampler`: assignment of SPT data points to microdomains
    * :attr:`~tramway.analyzer.RWAnalyzer.mapper`: estimation of model parameters at each microdomains

    Most attributes are self-morphing, i.e. they first are initializers and exhibit
    *from_...* methods (for example
    :meth:`~.spt_data.SPTDataInitializer.from_dataframe` and
    :meth:`~.spt_data.SPTDataInitializer.from_ascii_file` for
    :attr:`~tramway.analyzer.RWAnalyzer.spt_data`) and then,
    once any such initializer method is called,
    they specialize into a new attribute and exhibit specific attributes depending
    on the chosen initializer:

    .. code-block:: python

        from tramway.analyzer import *
        a = RWAnalyzer()
        a.spt_data.from_ascii_files('my_data_repository/*.txt')

    The more conventional assignment form is also available using functions
    of modules exported by the ``tramway.analyzer`` subpackage.
    These modules are named following the attribute they are dedicated to:

    .. code-block:: python

        a = RWAnalyzer()
        a.spt_data = spt_data.from_ascii_files('my_data_repository/*.txt')

    In most cases, a third form is also available, with a class constructor:

    .. code-block:: python

        a = RWAnalyzer()
        a.spt_data = spt_data.SPTAsciiFiles('my_data_repository/*.txt')

    Anyway, most of the attributes of an :class:`RWAnalyzer` object
    are properties that perform various checks and may raise exceptions
    if misused.

    Specialized attributes can also exhibit self-morphing attributes.
    For example, regions of interest can be defined globally using the main
    :attr:`~tramway.analyzer.RWAnalyzer.roi` attribute, or on a per-SPT-datafile basis:

    .. code-block:: python

        a = RWAnalyzer()
        a.spt_data.from_ascii_files('my_data_repository/*.txt')
        a.roi.from_squares(roi_centers, square_size)

    or

    .. code-block:: python

        a = RWAnalyzer()
        a.spt_data.from_ascii_files('my_data_repository/*.txt')
        for spt_file in a.spt_data:
            spt_file.roi.from_squares(roi_centers, square_size)

    In the above example, per-dataset ROI definition is useful when multiple
    datasets are loaded and the ROI may differ between datasets.
    The main :attr:`~tramway.analyzer.RWAnalyzer.roi` attribute is still convenient
    as it allows to iterate over all the defined ROI,
    omitting the :attr:`~tramway.analyzer.RWAnalyzer.spt_data` loop (continues any
    of the code blocks above):

    .. code-block:: python

        for r in a.roi:
            roi_spt_data = r.crop()

    The ROI object is denoted ``r`` here.
    Name ``roi`` already exists if the :mod:`~tramway.analyzer` module is
    imported with ``from tramway.analyzer import *``.
    In this case, ``roi`` points to the :mod:`~tramway.analyzer.roi` module.

    Note there exists multiple iterators for the collections of ROI.
    In the further examples of iterated ROI, the preferred iterator is made explicit.
    See the documentation for the :attr:`~tramway.analyzer.RWAnalyzer.roi` attribute
    for more information about the available iterators.

    While the :attr:`~tramway.analyzer.RWAnalyzer.spt_data` and
    :attr:`~tramway.analyzer.RWAnalyzer.roi` attributes act as data providers,
    the :attr:`~tramway.analyzer.RWAnalyzer.time`,
    :attr:`~tramway.analyzer.RWAnalyzer.tesseller`,
    :attr:`~tramway.analyzer.RWAnalyzer.sampler` and
    :attr:`~tramway.analyzer.RWAnalyzer.mapper` attributes
    do not feature direct access to the data and require the SPT data to be passed as
    input argument to their main processing methods.
    For example:

    .. code-block:: python

        a.tesseller.from_plugin('kmeans')
        for r in a.roi.as_support_regions():
            roi_spt_data = r.crop()
            tessellation = a.tesseller.tessellate(roi_spt_data)

    Similarly, the :attr:`~tramway.analyzer.RWAnalyzer.images` attribute defines data location,
    while the :attr:`~tramway.analyzer.RWAnalyzer.localizer` and
    :attr:`~tramway.analyzer.RWAnalyzer.tracker` attributes define processing steps
    on these data.

    Other attributes drive the execution of the processing chain.
    The :meth:`run` method launches the processing chain, which is operated by the
    :attr:`~tramway.analyzer.RWAnalyzer.pipeline` attribute.
    
    Various parallelization schemes are available, and the platform-specific
    implementation of these schemes are provided by the
    :attr:`~tramway.analyzer.RWAnalyzer.env` attribute.

    Last but not least, the :class:`RWAnalyzer` features plotting utilities.
    Some of them are available through the *mpl* sub-attribute of some
    main :class:`RWAnalyzer` attributes or items,
    for example :attr:`images.mpl` (:class:`~.images.mpl.Mpl`),
    :attr:`spt_data.mpl` (:class:`~.spt_data.mpl.Mpl`),
    :attr:`tesseller.mpl` (:class:`~.tesseller.mpl.Mpl`),
    :attr:`mapper.mpl` (:class:`~.mapper.mpl.Mpl`).
    In addition, the :attr:`~tramway.analyzer.RWAnalyzer.browser` attribute
    can plot the inferred parameter maps from a Jupyter notebook,
    or calling the ``bokeh serve`` command:

    .. code-block:: python

        from tramway.analyzer import *

        a = RWAnalyzer()

        # load the rwa files available in the current directory:
        a.spt_data.from_rwa_files('*.rwa')

        # help the analyzer locate this piece of code:
        try:
            a.script = __file__
        except NameError: # in a notebook
            a.script = 'MyNotebook.ipynb' # this notebook's name (please adapt)

        a.browser.show_maps()

    See also :class:`~tramway.analyzer.browser.Browser` for additional information
    on how to export data and figures while browsing the inferred parameter maps.

    """
    __slots__ = ( '_spt_data', '_roi', '_time', '_tesseller', '_sampler', '_mapper',
            '_env', '_pipeline', '_browser', '_images', '_localizer', '_tracker' )

    def _get_spt_data(self):
        """
        SPT data accessor.

        See :class:`~tramway.analyzer.spt_data.SPTDataInitializer`
        and :class:`~tramway.analyzer.spt_data.SPTData`.
        """
        return self._spt_data
    def _set_spt_data(self, data):
        self._spt_data = data
    spt_data = selfinitializing_property('spt_data', _get_spt_data, _set_spt_data, SPTData)

    def _get_roi(self):
        """
        ROI accessor.

        See :class:`~tramway.analyzer.roi.ROIInitializer`
        and :class:`~tramway.analyzer.roi.ROI`.
        """
        return self._roi
    def _set_roi(self, roi):
        self._roi = roi
    roi = selfinitializing_property('roi', _get_roi, _set_roi, ROI)

    def _get_time(self):
        """
        Time segmentation procedure.

        See :class:`~tramway.analyzer.time.TimeInitializer`
        and :class:`~tramway.analyzer.time.Time`.
        """
        return self._time
    def _set_time(self, time):
        self._time = time
    time = selfinitializing_property('time', _get_time, _set_time, Time)

    def _get_tesseller(self):
        """
        Tessellation procedure.

        See :class:`~tramway.analyzer.tesseller.TessellerInitializer`
        and :class:`~tramway.analyzer.tesseller.Tesseller`.
        """
        return self._tesseller
    def _set_tesseller(self, tesseller):
        self._tesseller = tesseller
    tesseller = selfinitializing_property('tesseller', _get_tesseller, _set_tesseller, Tesseller)

    def _get_sampler(self):
        """
        Sampling procedure.

        See :class:`~tramway.analyzer.sampler.SamplerInitializer`
        and :class:`~tramway.analyzer.sampler.Sampler`.
        """
        return self._sampler
    def _set_sampler(self, sampler):
        self._sampler = sampler
    sampler = selfinitializing_property('sampler', _get_sampler, _set_sampler, Sampler)

    def _get_mapper(self):
        """
        Inference procedure.

        See :class:`~tramway.analyzer.mapper.MapperInitializer`
        and :class:`~tramway.analyzer.mapper.Mapper`.
        """
        return self._mapper
    def _set_mapper(self, mapper):
        self._mapper = mapper
    mapper = selfinitializing_property('mapper', _get_mapper, _set_mapper, Mapper)

    def _get_env(self):
        """
        Environment backend for operating the pipeline.

        If not set, the pipeline will run locally in the current interpreter.

        See :mod:`~tramway.analyzer.env.environments`.
        """
        return self._env
    def _set_env(self, env):
        self._env = env
    env = selfinitializing_property('env', _get_env, _set_env, Environment)

    def _get_images(self):
        """
        Microscopy image stacks.

        See :class:`~tramway.analyzer.images.ImagesInitializer`
        and :class:`~tramway.analyzer.images.Images`.
        """
        return self._images
    def _set_images(self, images):
        self._images = images
    images = selfinitializing_property('images', _get_images, _set_images, Images)

    def _get_localizer(self):
        """
        Single molecule localization procedure.

        See :class:`~tramway.analyzer.localizer.LocalizerInitializer`
        and :class:`~tramway.analyzer.localizer.Localizer`.
        """
        return self._localizer
    def _set_localizer(self, localizer):
        self._localizer = localizer
    localizer = selfinitializing_property('localizer', _get_localizer, _set_localizer, Localizer)

    def _get_tracker(self):
        """
        Single particle tracking procedure.

        See :class:`~tramway.analyzer.tracker.TrackerInitializer`
        and :class:`~tramway.analyzer.tracker.Tracker`.
        """
        return self._tracker
    def _set_tracker(self, tracker):
        self._tracker = tracker
    tracker = selfinitializing_property('tracker', _get_tracker, _set_tracker, Tracker)

    def __init__(self):
        WithLogger.__init__(self)
        self._spt_data = \
                self._roi = \
                self._tesseller = \
                self._sampler = \
                self._mapper = \
                self._env = \
                self._images = \
                self._localizer = \
                self._tracker = None
        self.spt_data  = SPTDataInitializer
        self.roi       = ROIInitializer
        self.time      = TimeInitializer
        self.tesseller = TessellerInitializer
        self.sampler   = SamplerInitializer
        self.mapper    = MapperInitializer
        self.images    = ImagesInitializer
        self.localizer = LocalizerInitializer
        self.tracker   = TrackerInitializer
        self._pipeline = Pipeline(self)
        self.env       = EnvironmentInitializer
        self._browser  = Browser(self)

    @property
    def pipeline(self):
        """
        Parallelization scheme.

        See :class:`~tramway.analyzer.pipeline.Pipeline`.
        """
        return self._pipeline

    def run(self):
        """
        Launches the pipeline.

        Alias for :attr:`~tramway.analyzer.RWAnalyzer.pipeline` :meth:`~.pipeline.Pipeline.run`.
        """
        return self.pipeline.run()

    def add_collectible(self, collectible):
        """
        Designates a file generated on the worker side to be transferred back to the submit side.

        Alias for :attr:`~tramway.analyzer.RWAnalyzer.pipeline`
        :meth:`~.pipeline.Pipeline.add_collectible`.
        """
        self.pipeline.add_collectible(collectible)

    @property
    def browser(self):
        """
        Data visualization and export.

        See :class:`~.browser.Browser`.
        """
        return self._browser

    @property
    def script(self):
        """
        Path to the *__main__* file in which the analyzer is defined.

        Designating a script is required for parallelizing computations,
        or visualizing maps without explicitly calling the ``bokeh serve`` command
        (e.g. from a Jupyter notebook).

        Alias for :attr:`~tramway.analyzer.RWAnalyzer.env` :attr:`~.env.Environment.script`.
        """
        return self.env.script
    @script.setter
    def script(self, filename):
        self.env.script = filename

    def __setattr__(self, attrname, obj):
        if attrname[0] == '_' or (isinstance(obj, type) and issubclass(obj, Initializer)) or\
                attrname in ('script',):
            object.__setattr__(self, attrname, obj)
        elif callable(obj):
            if attrname[0] != '_' and isinstance(obj, InitializerMethod):
                attr = getattr(self, attrname)
                if isinstance(attr, Initializer):
                    obj.assign( self )
                else:
                    warnings.warn(
                            "attribute '{}' is already initialized; side effects may occur".format(
                                attrname),
                            SideEffectWarning)
                    obj.reassign( self )
            else:
                attr = getattr(self, attrname)
                try:
                    attr.from_callable(obj)
                except AttributeError as e:
                    if isinstance(attr, Initializer):
                        msg = f"shouldn't you call the value passed to attribute '{attrname}' beforehands?"
                        if not _emulate_undefined_from_callable:
                            raise ValueError(msg)
                        try:
                            setattr(self, attrname, obj() )
                        except (KeyboardInterrupt, SystemExit):
                            raise
                        except:
                            raise e
                        else:
                            warnings.warn(msg)
                    else:
                        raise AttributeError(f"attribute '{attrname}' is already initialized") from None
        else:
            parent_name, _, set_conditions = proper_parent_name(attrname)
            if parent_name is None:
                proper_parent = getattr(self, attrname) # raises AttributeError if no such attribute is found
                if proper_parent.initialized:
                    raise AttributeError('attribute `{}` is read-only'.format(attrname))
                else:
                    # new in 0.6: handle orphan specialized attribute
                    #             if not specialized yet
                    assert isinstance(proper_parent, Initializer)
                    obj._parent = proper_parent._parent
                    proper_parent._specialize(obj)
            else:
                report_misplaced_attribute(attrname, parent_name)
                proper_parent = getattr(self, parent_name)
                if set_conditions.get('initialized', False) and isinstance(proper_parent, Initializer):
                    raise AttributeError('`{}` cannot be set as long as `{}` is not initialized'.format(attrname, parent_name))
                setattr(proper_parent, attrname, obj)

    def __getattr__(self, attrname):
        parent_name, _, _ = proper_parent_name(attrname)
        if parent_name is None:
            raise AttributeError('RWAnalyzer has no attribute `{}`'.format(attrname))
        else:
            proper_parent = getattr(self, parent_name)
            return getattr(proper_parent, attrname)


__all__ = ['RWAnalyzer',
        'spt_data', 'roi',
        'time', 'tesseller', 'sampler',
        'mapper',
        'images', 'tracker',
        'tessellers', 'cell_mergers',
        'models',
        'Analysis', 'commit_as_analysis',
        'environments',
        'first', 'single',
        'SideEffectWarning',
        'stages',
        ]




# -*- coding: utf-8 -*-

# Copyright © 2017-2019, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


import numpy as np
import pandas as pd
import re
import sys
import collections

def isstructured(x):
    """
    Check for named columns.

    The adjective *structured* comes from NumPy structured array.

    Arguments:
        x (any): any datatype

    Returns:
        bool: ``True`` if input argument ``x`` has named columns.

    """
    if isinstance(x, (pd.Series, pd.DataFrame)):
        return True
    else:
        try:
            return bool(x.dtype.names)
        except AttributeError:
            return False

def columns(x):
    """
    Get column names.

    Arguments:
        x (any):
            datatype that satisfies :func:`isstructured`.

    Returns:
        iterable:
            column iterator.

    Raises:
        ValueError: if no named columns are found in `x`.

    """
    if isinstance(x, pd.DataFrame):
        return x.columns
    elif isinstance(x, pd.Series):
        return x.index
    elif x.dtype.names:
        return x.dtype.names
    else:
        raise ValueError('not structured')

def splitcoord(varnames, asstr=True):
    if asstr and sys.version_info[0] < 3:
        def _str(s):
            return s.encode('utf-8') if isinstance(s, unicode) else s
    else:
        _str = lambda s: s
    coord = re.compile('[a-z]?[0-9]*$')
    vs = collections.defaultdict(list)
    for v in varnames:
        u = [ w[::-1] for w in v[::-1].split(None, 1) ]
        if u[1:] and coord.match(u[0]):
            vs[_str(u[1])].append(v)
        else:
            vs[_str(v)].append(v)
    return vs

def expandcoord(varname, dim):
    if dim < 1:
        raise ValueError('`dim` is lower than 1')
    elif dim == 1:
        return [varname]
    elif dim <= 3:
        return [ '{} {}'.format(varname, c) for c in 'xyz'[:dim] ]
    else:
        return [ '{} x{:d}'.format(varname, c) for c in range(dim) ]


__all__ = [
    'isstructured',
    'columns',
    'splitcoord',
    'expandcoord',
    ]





from ..attribute import *

class Browser(AnalyzerNode):
    """
    :attr:`~tramway.analyzer.RWAnalyzer.browser` attribute of an
    :class:`~tramway.analyzer.RWAnalyzer` object.

    Rendering is based on *bokeh* and is performed in a web browser tab.
    An optional side panel with experimental export features is shown if
    argument ``side_panel=True`` is passed to :meth:`show_maps` or argument
    `webdriver` is defined:

    .. code-block:: python

        from tramway.analyzer import *

        a = RWAnalyzer()

        a.spt_data.from_rwa_files('*.rwa')

        from selenium import webdriver

        a.browser.show_maps(webdriver=webdriver.Firefox)

    Defining a webdriver is required for exporting figures.
    Supported file formats are *.png* and *.svg*.
    Note that package `selenium <https://selenium-python.readthedocs.io/installation.html>`_ is required.

    The above example can be explicitly run with ``bokeh serve``,
    or else it can be run with the standard Python interpreter or in a Jupyter notebook
    as long as the *script* attribute is defined:

    .. code-block:: python

        from tramway.analyzer import *

        a = RWAnalyzer()

        a.spt_data.from_rwa_files('*.rwa')

        try:
            a.script = __file__
        except NameError: # in a notebook
            a.script = 'MyNotebook.ipynb' # this notebook's name; please adapt

        from selenium import webdriver

        a.browser.show_maps(webdriver=webdriver.Firefox)

    The later example will call ``bokeh serve --show`` on the file specified in the *script* attribute.

    The showed parameter values can also be exported with the side panel.
    Note that all features are exported together with the spatial bin center coordinates.
    """
    __slots__ = ('_colormap','_clim','_this_notebook_cell_only')
    @property
    def colormap(self):
        """ *str*: Colormap for inferred parameter maps.

        See also :func:`~tramway.plot.bokeh.map.scalar_map_2d`.
        """
        return self._colormap
    @colormap.setter
    def colormap(self, cm):
        self._colormap = cm
    @property
    def clim(self):
        """
        *dict* of 2-element *array-like*:
            Color lower and upper values (dict values) for each feature (dict keys)
        """
        return self._clim
    @clim.setter
    def clim(self, clim):
        self._clim = clim
    def __init__(self, analyzer):
        AnalyzerNode.__init__(self, parent=analyzer)
        self._colormap = None
        self._clim = None
        self._this_notebook_cell_only = None
    @property
    def script(self):
        return self._parent.script
    @script.setter
    def script(self, path):
        self._parent.script = path
    @property
    def this_notebook_cell_only(self):
        return self._this_notebook_cell_only
    @this_notebook_cell_only.setter
    def this_notebook_cell_only(self, cell_index):
        self._this_notebook_cell_only = cell_index
    def show_maps(self, **kwargs):
        """ See also :func:`~tramway.plot.bokeh.analyzer.browse_maps`. """
        if self.this_notebook_cell_only and \
                self.script is not None and \
                self.script.endswith('.ipynb'):
            self.logger.info("""\
the designated .ipynb file will be exported;
be sure to save the notebook after any modification
so that the changes are reported in the .ipynb file
""")
            # export the .ipynb file to a Python script
            script_path = self.script
            script_content = import_ipynb(script_path)
            # discard the other code cells
            script_content = split_cells(script_content)
            if isinstance(self.this_notebook_cell_only, bool):
                caller_cells = []
                for _, cell in script_content:
                    for line in cell:
                        line = line.split('#')[0]
                        if '.browser.show_maps(' in line:
                            # a cell may appear several times
                            caller_cells.append(cell)
                if not caller_cells:
                    raise RuntimeError("cannot find an expression calling `browser.show_maps`")
                if caller_cells[1:]:
                    raise RuntimeError('multiple `browser.show_maps` calls found; please specify the index of the notebook cell')
                caller_cell = caller_cells[0]
            else:
                notebook_cell_index = self.this_notebook_cell_only
                try:
                    _, caller_cell = script_content[notebook_cell_index]
                except IndexError:
                    for i, cell in enumerate(script_content):
                        print(f'notebook cell {i}:\n')
                        print(''.join(cell[1]))
                    raise
                else:
                    self.logger.debug(f'notebook cell index {notebook_cell_index}:\n\n'+''.join(caller_cell))
            condensed_code = []
            for line in caller_cell:
                line = line.split('#')[0].strip()
                if line and \
                        'script' not in line and \
                        'this_notebook_cell_only' not in line:
                    condensed_code.append(line+'\n')
            # run bokeh serve on the exported notebook cell
            import tempfile, os, subprocess, sys
            f, tmpfile = tempfile.mkstemp(suffix='.py')
            try:
                for line in condensed_code:
                    os.write(f, line.encode('utf-8'))
                os.close(f)
                self.logger.info(f'running: python -m bokeh serve --show {tmpfile}')
                p = subprocess.Popen([sys.executable, '-m',
                        'bokeh', 'serve', '--show', tmpfile],
                    cwd=os.getcwd(),
                    stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                    close_fds=True,
                    encoding='utf-8')
                self.logger.info('interrupt the kernel to stop bokeh and proceed back in the notebook')
                try:
                    out, err = p.communicate()
                except KeyboardInterrupt:
                    out, err = p.communicate()
                if err:
                    self.logger.error(err)
                else:
                    if out:
                        self.logger.debug(out)
            finally:
                try:
                    os.unlink(tmpfile)
                except FileNotFoundError:
                    pass
        else:
            from tramway.plot.bokeh.analyzer import browse_maps
            browse_maps(self._eldest_parent, **kwargs)

Attribute.register(Browser)


def import_ipynb(path):
    from tramway.analyzer.env.environments import Env
    return Env().import_ipynb(path)

def split_cells(script):
    import re
    cells = []
    cell_index = None
    cell = []
    match = True
    for line in script:
        line_ = line.rstrip()
        match_with_index = re.fullmatch(r'# In\[(?P<cell_index>[1-9][0-9]*)\]:', line_)
        match_without_index = line_ == r'# In[ ]:'
        match = match_with_index or match_without_index
        if match:
            if cells or cell_index is not None or cell:
                cells.append((cell_index, cell))
            if match_with_index:
                cell_index = int(match_with_index.group('cell_index'))
            cell = []
        cell.append(line)
    if not match:
        cells.append((cell_index, cell))
    return cells


__all__ = ['Browser']





from . import *
from . import utils
from ..attribute import initializer_method

from_bounding_boxes = initializer_method( ROIInitializer.from_bounding_boxes )
from_squares        = initializer_method( ROIInitializer.from_squares        )
from_ascii_file     = initializer_method( ROIInitializer.from_ascii_file     )
from_ascii_files    = initializer_method( ROIInitializer.from_ascii_files    )
from_dedicated_rwa_record  = initializer_method( ROIInitializer.from_dedicated_rwa_record  )
from_dedicated_rwa_records = initializer_method( ROIInitializer.from_dedicated_rwa_records )




# -*- coding: utf-8 -*-

# Copyright © 2017-2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from . import abc
import itertools
import copy
import warnings


class AnalysesView(dict):
    __slots__ = ('__analyses__',)
    def __init__(self, analyses):
        self.__analyses__ = analyses
    def __nonzero__(self):
        #return self.__analyses__._instances.__nonzero__()
        return bool(self.__analyses__._instances) # Py2
    def __len__(self):
        return len(self.__analyses__._instances)
    def __missing__(self, label):
        raise KeyError('no such analysis instance: {}'.format(label))
    def __iter__(self):
        return self.__analyses__._instances.__iter__()
    def __contains__(self, label):
        return self.__analyses__._instances.__contains__(label)
    def keys(self):
        return self.__analyses__._instances.keys()

class InstancesView(AnalysesView):
    __slots__ = ()
    def __str__(self):
        return self.__analyses__._instances.__str__()
    def __getitem__(self, label):
        return self.__analyses__._instances[label]
    def __setitem__(self, label, analysis):
        if not isinstance(analysis, abc.Analyses):
            analysis = type(self.__analyses__)(analysis)
        self.__analyses__._instances[label] = analysis
    def __delitem__(self, label):
        self.__analyses__._instances.__delitem__(label)
        try:
            self.__analyses__._comments.__delitem__(label)
        except KeyError:
            pass
    def values(self):
        return self.__analyses__._instances.values()
    def items(self):
        return self.__analyses__._instances.items()
    def get(self, label, default=None):
        return self.__analyses__._instances.get(label, default)
    def pop(self, label, default=None):
        analysis = self.get(label, default)
        self.__delitem__(label)
        return analysis

class CommentsView(AnalysesView):
    __slots__ = ()
    def __str__(self):
        return self.__analyses__._comments.__str__()
    def __getitem__(self, label):
        try:
            return self.__analyses__._comments[label]
        except KeyError:
            if label in self.__analyses__._instances:
                return None
            else:
                self.__missing__(label)
    def __setitem__(self, label, comment):
        if label in self.__analyses__._instances:
            if comment:
                self.__analyses__._comments[label] = comment
            else:
                self.__delitem__(label)
        else:
            self.__missing__(label)
    def __delitem__(self, label):
        self.__analyses__._comments.__delitem__(label)
    def values(self):
        return self.__analyses__._comments.values()
    def items(self):
        return self.__analyses__._comments.items()


class Analyses(object):
    """
    Analysis tree - Generic container with labels, comments and other metadata to structure
    the analyses that apply to the same data.

    An :class:`Analyses` object is a node of a tree.
    In attribute `data` (or equivalently `artefact`)
    it contains the input data for the children analyses,
    and these children analyses can be accessed as subtrees using a dict-like interface.

    Labels of the children analyses can be listed with property `labels`.
    A label is a key in the dict-like interface.

    Comments associated to children analyses are also addressable with labels.

    Metadata are attached to each node, including the top node.

    Setting the `data` attribute unsets the other attributes.

    Example:

    .. code-block:: python

        ## let `my_input_data` and `my_output_data` be dataframes:
        #my_output_data = my_analysis(my_input_data)

        ## build the tree
        tree = Analyses(my_input_data) # root node

        tree.add(my_output_data, label='my analysis', comment='description of my analysis')
        # or equivalently (order matters):
        tree['my analysis'] = my_output_data
        tree.comments['my analysis'] = 'description of my analysis'

        ## print
        print(tree)
        #<class 'pandas.core.frame.DataFrame'>
        #        'my analysis' <class 'pandas.core.frame.DataFrame'>:    "description of my analysis"

        assert tree.data is my_input_data
        # note that `my_output_data` has been automatically wrapped into an `Analysis` object:
        assert isinstance(tree['my analysis'], Analyses)
        assert tree['my analysis'].data is my_output_data

        print(tree.labels) # or print(tree.keys())
        #dict_keys(['my analysis'])

        print(tree.comments['my analysis'])
        #description of my analysis


    Attributes:

        data/artefact (any): input data to the children instances.

        instances (dict): analyses on the data; keys are natural integers or string labels.

        comments (dict): comments associated to the analyses; keys are a subset of the keys
            in `instances`.

        metadata (dict): additional metadata associated to the input data; keys are attributes
            and are not related to children instances.

    """
    __slots__ = ('_data', '_instances', '_comments', '_metadata')

    def __init__(self, data=None, metadata=None):
        self._data = data
        self._instances = {}
        self._comments = {}
        self._metadata = {} if metadata is None else metadata

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, d):
        self._data = d
        self._instances = {}
        self._comments = {}
        self._metadata = {}

    @property
    def artefact(self):
        return self.data

    @artefact.setter
    def artefact(self, a):
        self.data = a

    @property
    def metadata(self):
        if self._metadata is None:
            self._metadata = {}
        return self._metadata

    @metadata.setter
    def metadata(self, d):
        self._metadata = {} if d is None else d

    @property
    def instances(self):
        return InstancesView(self)

    @property
    def comments(self):
        return CommentsView(self)

    @property
    def labels(self):
        return self.instances.keys()

    def keys(self):
        return self.instances.keys()

    def autoindex(self, pattern=None):
        r"""
        Determine the lowest available natural integer for use as key in `instances` and `comments`.

        If `pattern` is an integer, `autoindex` returns the pattern unchanged.

        Arguments:

            pattern (str): label with a *'\*'* to be replaced by a natural integer.

        Returns:

            int or str: index or label.
        """
        if valid_label(pattern):
            try:
                pattern = int(pattern) # this also checks type
            except ValueError:
                pass
            if not isinstance(pattern, int) and '*' in pattern:
                f = lambda i: pattern.replace('*', str(i))
            else:
                return pattern
        else:
            f = lambda i: i
        i = 0
        if self.instances:
            while f(i) in self.instances:
                i += 1
        return f(i)

    def add(self, analysis, label=None, comment=None, raw=False):
        """
        Add an analysis.

        Adding an analysis at an existing label overwrites the existing analysis
        instance and deletes the associated comment if any.

        Arguments:

            analysis (any): analysis instance.

            label (any): key for the analysis; calls :meth:`autoindex` if undefined.

            comment (str): associated comment.

            raw (bool):
                if `analysis` is not an :class:`~tramway.core.analyses.base.Analyses`,
                it is wrapped into such a container object;
                set `raw` to ``True`` to prevent wrapping.

        """
        label = self.autoindex(label)
        if not (raw or isinstance(analysis, abc.Analyses)):
            analysis = type(self)(analysis)
        self.instances[label] = analysis
        if comment:
            self.comments[label] = comment
        else:
            try:
                del self.comments[label]
            except KeyError:
                pass

    def __nonzero__(self):
        return self.instances.__nonzero__()
    def __len__(self):
        return self.instances.__len__()
    def __missing__(self, label):
        self.instances.__missing__(label)
    def __iter__(self):
        return self.instances.__iter__()
    def __contains__(self, label):
        return self.instances.__contains__(label)
    def __getitem__(self, label):
        return self.instances.__getitem__(label)
    def __setitem__(self, label, analysis):
        self.instances.__setitem__(label, analysis)
    def __delitem__(self, label):
        self.instances.__delitem__(label)

abc.Analyses.register(Analyses)


valid_label = lambda label: label or (isinstance(label, int) and label == 0)


def map_analyses(fun, analyses, label=False, comment=False, metadata=False, depth=False,
        allow_tuples=False):
    with_label = label or valid_label(label)
    with_comment, with_metadata, with_depth = comment, metadata, depth
    def _fun(x, **kwargs):
        y = fun(x, **kwargs)
        if not allow_tuples and isinstance(y, tuple):
            raise ValueError('type conflict: function returned a tuple')
        return y
    def _map(analyses, label=None, comment=None, depth=0):
        kwargs = {}
        if with_label:
            kwargs['label'] = label
        if with_comment:
            kwargs['comment'] = comment
        if with_metadata:
            kwargs['metadata'] = analyses.metadata
        if with_depth:
            kwargs['depth'] = depth
        node = _fun(analyses._data, **kwargs)
        if analyses.instances:
            depth += 1
            tree = []
            for label in analyses.instances:
                child = analyses.instances[label]
                comment = analyses.comments[label]
                if isinstance(child, abc.Analyses):
                    tree.append(_map(child, label, comment, depth))
                else:
                    if with_label:
                        kwargs['label'] = label
                    if with_comment:
                        kwargs['comment'] = comment
                    if with_depth:
                        kwargs['depth'] = depth
                    tree.append(_fun(child, **kwargs))
            return node, tuple(tree)
        else:
            return node
    return _map(analyses)


def extract_analysis(analyses, labels):
    """
    Extract an analysis from a hierarchy of analyses.

    The elements of an :class:`~tramway.core.analyses.base.Analyses` instance can be other
    :class:`~tramway.core.analyses.base.Analyses` objects.
    As such, analyses are structured in a tree that exhibits as many logically-consecutive
    layers as there are processing steps.

    Arguments:

        analyses (tramway.core.analyses.base.Analyses):
            hierarchy of analyses, with `instances` possibly containing
            other :class:`~tramway.core.analyses.base.Analyses` instances.

        labels (int, str or sequence of int and str):
            analyses label(s); the first label addresses the first layer of
            analyses instances, the second label addresses the second layer of
            analyses and so on.

    Returns:

        tramway.core.analyses.base.Analyses: copy of the analyses along the path defined by `labels`.
    """
    analysis = type(analyses)(analyses._data, analyses.metadata)
    if labels:
        if not isinstance(labels, (tuple, list)):
            labels = [labels]
        label = labels[0]
        analysis.instances[label] = extract_analyses(analyses.instances[label], labels[1:])
        try:
            analysis.comments[label] = analyses.comments[label]
        except KeyError:
            pass
    return analysis


def _append(s, ls):
    """for internal use in :fun:`label_paths`."""
    if ls:
        ss = []
        for ok, _ls in ls:
            if isinstance(ok, bool):
                l, _ls = _ls, None
            else:
                ok, l = ok
            _s = list(s) # copy
            _s.append(l)
            if ok:
                ss.append([tuple(_s)])
            if _ls:
                ss.append(_append(_s, _ls))
        return itertools.chain(*ss)
    else:
        return []

def label_paths(analyses, filter):
    """
    Find label paths for analyses matching a criterion.

    Arguments:

        analyses (tramway.core.analyses.base.Analyses):
            hierarchy of analyses, with `instances` possibly containing
            other :class:`~tramway.core.analyses.base.Analyses` instances.

        filter (type or callable):
            criterion over analysis data.

    Returns:

        list of tuples:
            list of label paths to matching analyses.

    """
    if isinstance(filter, type):
        _type = filter
        _map = lambda node, label: (isinstance(node, _type), label)
    elif callable(filter):
        _map = lambda node, label: (filter(node), label)
    else:
        raise TypeError('`filter` is neither a type nor a callable')
    _, labels = map_analyses(_map, analyses, label=True, allow_tuples=True)
    return list(_append([], labels))


def find_artefacts(analyses, filters, labels=None, quantifiers=None, fullnode=False,
        return_subtree=False):
    """
    Find related artefacts.

    Filters are applied to find data elements (artefacts) along a single path specified by `labels`.

    Arguments:

        analyses (tramway.core.analyses.base.Analyses): hierarchy of analyses.

        filters (type or callable or tuple or list): list of criteria, a criterion being
            a boolean function or a type.

        labels (list): label path.

        quantifiers (str or tuple or list): list of quantifers, a quantifier for now being
            either '*first*', '*last*' or '*all*'; a quantifier should be defined for each
            filter; default is '*last*' (admits value ``None``).

        return_subtree (bool): return as extra output argument the analysis subtree corresponding
            to the deepest matching artefact.

    Returns:

        tuple: matching data elements/artefacts, and optionally analysis subtree.

    Examples:

    .. code-block:: python

        cells, maps = find_artefacts(analyses, (CellStats, Maps))

        maps, maps_subtree = find_artefacts(analyses, Maps, return_subtree=True)

    """
    # filters
    if not isinstance(filters, (tuple, list)):
        filters = (filters,)
    # quantifiers
    quantifier = 'last'
    if quantifiers:
        if not isinstance(quantifiers, (tuple, list)):
            quantifiers = (quantifiers,)
        if len(quantifiers) == len(filters):
            filters = zip(quantifiers, filters)
        elif quantifiers[1:]:
            warnings.warn('wrong number of quantifiers; ignoring them')
        else:
            quantifier = quantifiers[0]
    # labels
    if labels is None:
        labels = []
    elif isinstance(labels, (tuple, list)):
        labels = list(labels) # copy
    else:
        labels = [labels]
    labels_defined = bool(labels)
    if labels_defined:
        labels = labels[::-1]
    subtree, matches, lookup = None, [], True
    for i, _filter in enumerate(filters):
        if quantifiers:
            quantifier, _filter = _filter
        if isinstance(_filter, (type, tuple, list)):
            _type = _filter
            _filter = lambda a: isinstance(a.data, _type)
        elif callable(_filter):
            if not fullnode:
                f = _filter
                _fitler = lambda a: f(a.data)
        else:
            raise TypeError('invalid filter type: {}'.format(type(_filter)))
        match = []
        while True:
            if lookup:
                if labels:
                    label = labels.pop()
                elif labels_defined:
                    if match and i + 1 == len(filters):
                        break
                    raise ValueError('no match for {}{} filter'.format(i+1,
                        {1: 'st', 2: 'nd', 3: 'rd'}.get(i+1, 'th')))
                else:
                    _labels = list(analyses.labels)
                    if _labels and not _labels[1:]:
                        label = _labels[0]
                    elif match and i + 1 == len(filters):
                        break
                    elif not _labels:
                        raise ValueError('no match for {}{} filter'.format(i+1,
                            {1: 'st', 2: 'nd', 3: 'rd'}.get(i+1, 'th')))
                    else:#if _labels[1:]:
                        raise ValueError('multiple labels; argument `labels` required')
                try:
                    analyses = analyses.instances[label]
                except KeyError:
                    available = str(list(analyses.labels))[1:-1]
                    if available:
                        raise KeyError("missing label '{}'; available labels are: {}".format(label, available))
                    else:
                        raise KeyError("missing label '{}'; no labels available".format(label))
            lookup = True
            if _filter(analyses):
                match.append(analyses)
            elif match:
                lookup = False
                break
        if quantifier in ('first', ):
            subtree = match[0]
            match = subtree._data
        elif quantifier in ('last', None):
            subtree = match[-1]
            match = subtree._data
        elif quantifier in ('all', '+'):
            subtree = match
            match = [ a._data for a in match ]
        else:
            raise ValueError('invalid quantifier: {}'.format(quantifier))
        matches.append(match)
    if return_subtree:
        matches.append(subtree)
    return tuple(matches)


def coerce_labels_and_metadata(analyses):
    for key in analyses.metadata:
        val = analyses.metadata[key]
        if not isinstance(val, str):
            if isinstance(val, bytes):
                val = val.decode('utf-8')
            elif isinstance(val, unicode):
                val = val.encode('utf-8')
        analyses.metadata[key] = val
    for label in tuple(analyses.labels):
        if isinstance(label, (int, str)):
            coerced = label
        else:
            try: # Py2
                coerced = label.encode('utf-8')
            except AttributeError: # Py3
                try:
                    coerced = label.decode('utf-8')
                except AttributeError: # numpy.int64?
                    coerced = int(label)
            assert isinstance(coerced, (int, str))
        comment = analyses.comments[label]
        analysis = analyses.instances.pop(label)
        if isinstance(analysis, abc.Analyses):
            analysis = coerce_labels_and_metadata(analysis)
        analyses.instances[coerced] = analysis
        if comment:
            analyses.comments[coerced] = comment
    return analyses

coerce_labels = coerce_labels_and_metadata


def format_analyses(analyses, prefix='\t', node=type, global_prefix='', format_standalone_root=None,
        metadata=False, annotations={}):
    if isinstance(annotations, dict):
        annotations = dict(annotations) # copy
        root_annotation = annotations.get('', None)
    elif callable(annotations):
        root_annotation = annotations
    elif annotations:
        root_annotation = annotations.pop(0)
    if format_standalone_root is None:
        if metadata:
            if root_annotation is None:
                format_standalone_root = lambda r: '<Analyses {}>\n{}'.format(*r.split('\n',1))
            elif callable(root_annotation):
                def _format_standalone_root(r):
                    _type, _metadata = r.split('\n',1)
                    _annotation = root_annotation('', _type)
                    if _annotation:
                        return '<Analyses {}>\t<- {}\n{}'.format(_type, _annotation, _metadata)
                    else:
                        return '<Analyses {}>\n{}'.format(*r.split('\n',1))
                format_standalone_root = _format_standalone_root
            else:
                format_standalone_root = lambda r: '<Analyses {1}>\t<- {0}\n{2}'.format(root_annotation, *r.split('\n',1))
        else:
            if root_annotation is None:
                format_standalone_root = lambda r: '<Analyses {}>'.format(r)
            else:
                format_standalone_root = lambda r: '<Analyses {}>\t<- {}'.format(r, root_annotation)
    def _format(data, label=None, comment=None, metadata=None, depth=0):
        _prefix = global_prefix + prefix * depth
        s = [_prefix]
        t = []
        if label is None:
            assert comment is None
            if node:
                _node = node(data)
                desc = str(_node)
                s.append(desc)
            else:
                return None
        else:
            try:
                label + 0 # check numeric types
            except TypeError:
                s.append("'{}'")
            else:
                s.append('[{}]')
            t.append(label)
            if node:
                s.append(' {}')
                desc = _node = node(data)
                t.append(desc)
            if isinstance(annotations, dict):
                annotation = annotations.get(label, None)
            elif callable(annotations):
                annotation = annotations(label, _node)
            elif annotations:
                annotation = annotations.pop(0)
            if annotation is not None:
                s.append('\t<- {}')
                t.append(annotation)
            if comment:
                assert isinstance(comment, str)
                _comment = comment.split('\n')
                s.append('\n{}"{}')
                t.append(_prefix)
                t.append(_comment[0])
                for _line in _comment[1:]:
                    s.append('\n{} {}')
                    t.append(_prefix)
                    t.append(_line)
                s.append('"')
        if metadata:
            for key in metadata:
                val = metadata[key]
                s.append('\n{}@{}={}')
                t.append(_prefix)
                t.append(key)
                t.append(val)
        return ''.join(s).format(*t)
    def _flatten(_node):
        if _node is None:
            return []
        elif isinstance(_node, str):
            return [ _node ]
        try:
            _node, _children = _node
        except TypeError:
            return []
        else:
            assert isinstance(_node, str)
            return itertools.chain([_node], *[_flatten(c) for c in _children])
    entries = list(_flatten(map_analyses(_format, analyses, label=True, comment=True, metadata=metadata, depth=True)))
    if entries[1:]:
        if root_annotation:
            r = entries[0]
            if metadata:
                _type, _metadata = r.split('\n',1)
            else:
                _type = r
            if callable(root_annotation):
                _annotation = root_annotation('', _type)
            else:
                _annotation = root_annotation
            if _annotation:
                it = iter(entries[1:])
                _min_offset = None
                while _min_offset is None:
                    try:
                        s = next(it)
                        _min_offset = s.index('<-')
                    except ValueError:
                        pass
                _ntabs = s.count('\t', 0, _min_offset)
                _tab = lambda s: ' '*(_min_offset-_ntabs-len(s))+'\t'*_ntabs
                if metadata:
                    entries[0] = '{}{}<- {}\n{}'.format(_type, _tab(_type), _annotation, _metadata)
                else:
                    entries[0] = '{}{}<- {}'.format(r, _tab(r), _annotation)
        return '\n'.join(entries)
    else:
        return format_standalone_root(entries[0])


def append_leaf(analysis_tree, augmented_branch, overwrite=False):
    """
    Merge new analyses into an existing analysis tree.

    Only leaves and missing branches are appended.
    Existing nodes with children nodes are left untouched.

    Arguments:

        analysis_tree (tramway.core.analyses.base.Analyses): existing analysis tree.

        augmented_branch (tramway.core.analyses.base.Analyses): sub-tree with extra leaves.

    """
    if augmented_branch:
        for label in augmented_branch:
            if label in analysis_tree and (not overwrite or augmented_branch[label]):
                append_leaf(analysis_tree[label], augmented_branch[label])
            else:
                analysis_tree.add(augmented_branch[label], label=label)
    else:
        if analysis_tree:
            return
            #raise ValueError('the existing analysis tree has higher branches than the augmented branch')
        analysis_tree.data = augmented_branch.data



__all__ = [
    'AnalysesView',
    'InstancesView',
    'CommentsView',
    'Analyses',
    'map_analyses',
    'extract_analysis',
    'label_paths',
    'find_artefacts',
    'coerce_labels_and_metadata',
    'coerce_labels',
    'format_analyses',
    'append_leaf',
    'valid_label',
    ]

